###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Example of a DaVinci job filling all available functors. This is obviously a stress test and not realistic.

 This example is meant to be run with
    $ ./run davinci run-mc --inputfiledb Spruce_all_lines_dst Phys/DaVinci/options/DaVinciDB-Example.yaml --joboptfile DaVinciExamples/python/DaVinciExamples/tupling/option_davinci_tupling_from_spruce.yaml --user_algorithms  DaVinciExamples/python/DaVinciExamples/tupling/example-tupling-AllFunctors:alg_config --evt_max 100 |& cat | tee log
"""
__author__ = "P. Koppenburg"
__date__ = "2021-11-23"

import Functors as F
from FunTuple import FunctorCollection
from FunTuple import FunTuple_Particles as Funtuple
from PyConf.application import make_data_with_FetchDataFromFile
from DaVinci.reco_objects import make_pvs_for
from DaVinci.algorithms import add_filter
from DaVinci import options
from DecayTreeFitter import DTFAlg

#
# Definition of strucing line
#
bd2dsk_line = "SpruceB2OC_BdToDsmK_DsmToHHH_FEST_Line"
bd2dsk_data = make_data_with_FetchDataFromFile(
    f"/Event/Spruce/{bd2dsk_line}/Particles")

_basic = 'basic'
_composite = 'composite'

options.ntuple_file = "DV_example_allFunctors_ntp.root"
options.histo_file = "DV_example_allFunctors_his.root"


def all_variables(pvs, DTFR, ptype):
    """
    function that returns dictonary of functors that work.

    functors are listed in order of https://gitlab.cern.ch/lhcb/Rec/-/blob/master/Phys/FunctorCore/python/Functors/__init__.py
    """
    if ptype not in [_basic, _composite]:
        Exception("I want {0} or {1}. Got {2}".format(_basic, _composite,
                                                      ptype))
    all_vars = {}

    all_vars['REFERENCEPOINT_X'] = F.REFERENCEPOINT_X
    all_vars['REFERENCEPOINT_Y'] = F.REFERENCEPOINT_Y
    all_vars['REFERENCEPOINT_Z'] = F.REFERENCEPOINT_Z
    all_vars['TX'] = F.TX
    all_vars['TY'] = F.TY
    # all_vars['COV'] = F.COV # 'Track__Covariance' object has no attribute 'to_json'
    all_vars['P'] = F.P
    all_vars['PT'] = F.PT
    all_vars['PHI'] = F.PHI
    all_vars['ETA'] = F.ETA
    all_vars['PX'] = F.PX
    all_vars['PY'] = F.PY
    all_vars['PZ'] = F.PZ
    all_vars['ENERGY'] = F.ENERGY
    all_vars['FOURMOMENTUM'] = F.FOURMOMENTUM
    all_vars['MASS'] = F.MASS

    if (_basic == ptype):
        all_vars['ISMUON'] = F.ISMUON


#    all_vars['NDOF'] = F.NDOF # no member named 'nDoF' in 'LHCb::Particle
    all_vars['CHI2DOF'] = F.CHI2DOF
    all_vars['CHI2'] = F.CHI2

    if (_basic == ptype):
        all_vars['GHOSTPROB'] = F.GHOSTPROB
    #    all_vars['CLOSESTTOBEAM'] = F.CLOSESTTOBEAM # 'Track__ClosestToBeamState' object has no attribute 'to_json'
    #    all_vars['NHITS'] = F.NHITS  # 'no member named 'nHits' in 'LHCb::Particle''
    #    all_vars['NVPHITS'] = F.NVPHITS
    #    all_vars['NUTHITS'] = F.NUTHITS
    #    all_vars['NFTHITS'] = F.NFTHITS
    #    all_vars['TRACKHISTORY'] = F.TRACKHISTORY

    all_vars['MINIP'] = F.MINIP(pvs)
    all_vars['MINIPCHI2'] = F.MINIPCHI2(pvs)

    if (_composite == ptype):
        all_vars['SUMPT'] = F.SUM(F.PT)
        all_vars['MINPT'] = F.MIN(F.PT)
        all_vars['MAXPT'] = F.MAX(F.PT)
        all_vars['CHILD'] = F.CHILD(1, F.PT)
        all_vars['SUBCOMB12_MM'] = F.SUBCOMB(Functor=F.MASS, Indices=(1, 2))
        all_vars['MASSWITHHYPOTHESES'] = F.MASSWITHHYPOTHESES((939., 939.))
        all_vars['END_VX'] = F.END_VX
        all_vars['END_VY'] = F.END_VY
        all_vars['END_VZ'] = F.END_VZ
        all_vars['END_VRho'] = F.END_VRho
        all_vars['Ds_END_VZ'] = F.CHILD(1, F.END_VZ)
        all_vars['Delta_END_VZ_DsB0'] = F.CHILD(1, F.END_VZ) - F.END_VZ

        # use 'SDOCA' instad of 'DOCA' to avoid accessing the conditions DB for the magnetic field
        all_vars['DOCA'] = F.SDOCA(Child1=1, Child2=2)
        all_vars['DOCACHI2'] = F.SDOCACHI2(Child1=1, Child2=2)
        all_vars['ALV'] = F.ALV(Child1=1, Child2=2)
        all_vars['MAXDOCA'] = F.MAXSDOCA
        all_vars['MAXDOCACHI2'] = F.MAXSDOCACHI2

    all_vars['CHARGE'] = F.CHARGE

    if (_basic == ptype):
        all_vars['PID_MU'] = F.PID_MU
        all_vars['PID_PI'] = F.PID_PI
        all_vars['PID_K'] = F.PID_K
        all_vars['PID_P'] = F.PID_P
        all_vars['PID_E'] = F.PID_E
        all_vars['PROBNN_D'] = F.PROBNN_D
        all_vars['PROBNN_E'] = F.PROBNN_E
        all_vars['PROBNN_GHOST'] = F.PROBNN_GHOST
        all_vars['PROBNN_K'] = F.PROBNN_K
        all_vars['PROBNN_MU'] = F.PROBNN_MU
        all_vars['PROBNN_P'] = F.PROBNN_P
        all_vars['PROBNN_PI'] = F.PROBNN_PI

    #    all_vars['SIZE'] = F.SIZE(DTFR)  # no matching function for call to 'invoke'

    all_vars['BPVIP'] = F.BPVIP(pvs)
    all_vars['BPVIPCHI2'] = F.BPVIPCHI2(pvs)
    if (_composite == ptype):  # all these require a vertex
        all_vars['BPVETA'] = F.BPVETA(pvs)
        all_vars['BPVCORRM'] = F.BPVCORRM(pvs)
        all_vars['BPVDIRA'] = F.BPVDIRA(pvs)
        all_vars['BPVFDCHI2'] = F.BPVFDCHI2(pvs)
        all_vars['BPVVDZ'] = F.BPVVDZ(pvs)
        all_vars['BPVVDRHO'] = F.BPVVDRHO(pvs)
        all_vars['BPVLTIME'] = F.BPVLTIME(pvs)
        all_vars['BPVDLS'] = F.BPVDLS(pvs)
    #    all_vars['RUNNUMBER'] = F.RUNNUMBER(ODINLocation)    # doesn't work.
    #    all_vars['EVENTNUMBER'] = F.EVENTNUMBER('/Event/DAQ/RawBanks/ODIN')
    #    all_vars['EVENTTYPE'] = F.EVENTTYPE('/Event/DAQ/RawBanks/ODIN')
    #    all_vars['MVA'] = F.MVA    # needs an MVA
    #    all_vars['COMB'] = F.COMB  # starts from combination
    #    all_vars['POD'] = F.POD    # starts from combination

    all_vars['DTF_PT'] = F.MAP_INPUT(Functor=F.PT, Relations=DTFR)
    all_vars['DTF_BPVIPCHI2'] = F.MAP_INPUT(
        Functor=F.BPVIPCHI2(pvs), Relations=DTFR)

    print("For {0} returning variables {1}".format(ptype, all_vars.keys()))
    return all_vars


def alg_config():
    """
    Algorithm configuration function called from the comad line
    """
    #
    # DecayTreeFitter Algorithm
    #
    v2_pvs = make_pvs_for(process=options.process, data_type=options.data_type)

    #
    # DecayTreeFitter Algorithm
    #
    DTF = DTFAlg(Input=bd2dsk_data)
    # DTFParts = DTF.Output  # Particles (not needed)
    DTFRelations = DTF.OutputRelations  # Relations

    #
    # Definition of fields (branches) and functors
    #
    fields_dsk = {
        'B0': "[B0 -> D_s- K+]CC",
        'Kaon': "[B0 -> D_s- ^K+]CC",
    }
    variables_dsk = {
        'B0': FunctorCollection(
            all_variables(v2_pvs, DTFRelations, _composite)),
        'Kaon': FunctorCollection(all_variables(v2_pvs, DTFRelations, _basic)),
    }

    #
    # Sprucing filter
    #
    my_filter = add_filter("HDRFilter_B0DsK",
                           f"HLT_PASS('{bd2dsk_line}Decision')")

    #
    # FunTuple
    #
    my_tuple = Funtuple(
        name="B0DsK_Tuple",
        tuple_name="DecayTree",
        fields=fields_dsk,
        variables=variables_dsk,
        loki_preamble=[],
        inputs=bd2dsk_data)

    #
    # Algorithms to be run
    #
    return {"UserAlgs": [my_filter, my_tuple]}, []
