###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Example of a typical DaVinci job:
 - selection of two detached opposite-charge muons
 - tuple of the selected candidates
 - filter to select only the particles of interest
 - runs DecayTreeFitterAlg and stores some output

 This example is meant to be run with
    $ ./run davinci run-mc --inputfiledb Spruce_all_lines_dst Phys/DaVinci/options/DaVinciDB-Example.yaml --joboptfile DaVinciExamples/python/DaVinciExamples/tupling/example-tupling-DTF-filtered.yaml --user_algorithms DaVinciExamples/python/DaVinciExamples/tupling/example-tupling-DTF-filtered:main | tee log
"""

import Functors as F
from Gaudi.Configuration import INFO
from DaVinci.algorithms import filter_on, add_filter
from DecayTreeFitter import DTFAlg
from FunTuple import FunctorCollection
from FunTuple import FunTuple_Particles as Funtuple

#fields for FunTuple
fields = {}
fields['B0'] = "[B0 -> D_s- K+]CC"
fields['Ds'] = "[B0 -> ^D_s- K+]CC"
fields['K'] = "[B0 -> D_s- ^K+]CC"

#Get filtered particles (Note decay_descriptor is optional, if specified only B0 decays will be selected for processing)
spruce_line = "SpruceB2OC_BdToDsmK_DsmToHHH_FEST_Line"
data_filtered = filter_on(
    f"/Event/Spruce/{spruce_line}/Particles", decay_descriptor=fields['B0'])

# DecayTreeFitter Algorithm.
DTF = DTFAlg(Input=data_filtered, MassConstraints=["D_s-"], OutputLevel=INFO)
DTFRelations = DTF.OutputRelations  # Relations

#make collection of functors for all particles
variables_all = FunctorCollection({
    'THOR_P': F.P,
    'THOR_PT': F.PT,
    'THOR_MASS': F.MASS,
})

#make collection of functors for Ds meson
variables_ds = FunctorCollection({
    'DTF_PT':
    F.MAP_INPUT(Functor=F.PT, Relations=DTFRelations),
    'DTF_MASS':
    F.MAP_INPUT(Functor=F.MASS, Relations=DTFRelations),
})

#associate FunctorCollection to field (branch) name
variables = {'ALL': variables_all, 'Ds': variables_ds}

filter_data = add_filter("SpruceFilter", f"HLT_PASS('{spruce_line}Decision')")

#Configure Funtuple algorithm
tuple_data = Funtuple(
    name="B0DsK_Tuple",
    tuple_name="DecayTree",
    fields=fields,
    variables=variables,
    inputs=data_filtered)


def main():
    tools = []
    algs = [filter_data, tuple_data]
    return algs, tools
