###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Example of a typical DaVinci job:
 - selection of two detached opposite-charge muons
 - tuple of the selected candidates
 - runs DecayTreeFitterAlg and stores some output

 This example is meant to be run with
    $ ./run davinci run-mc --inputfiledb Upgrade_Bd2KstarMuMu_ldst Phys/DaVinci/options/DaVinciDB-Example.yaml --joboptfile DaVinciExamples/python/DaVinciExamples/tupling/example-tupling-DTF-run-mc.yaml --user_algorithms DaVinciExamples/python/DaVinciExamples/tupling/example-tupling-DTF-run-mc:main --evt_max -1 |& cat | tee log

"""

__author__ = "P. Koppenburg"
__date__ = "2021-11-23"

import Functors as F
from DaVinci.standard_particles import make_detached_mumu
from DaVinci.reco_objects import upfront_reconstruction_from_file as upfront_reconstruction
from DaVinci.reco_objects import make_pvs
from FunTuple import FunctorCollection
from FunTuple import FunTuple_Particles as Funtuple
from DecayTreeFitter import DTFAlg, DTF_functors

# Prepare the node with the selection
dimuons = make_detached_mumu()
pvs = make_pvs()

#
# DecayTreeFitter Algorithm.
# One with PV constraint and one without
#

DTF = DTFAlg(Input=dimuons, MassConstraints=["J/psi(1S)"], OutputLevel=3)
# DTFParts = DTF.Output  # Particles
DTFRelations = DTF.OutputRelations  # Relations

#FunTuple: Jpsi info
fields = {}
fields['Jpsi'] = 'J/psi(1S) -> mu+ mu-'
fields['MuPlus'] = 'J/psi(1S) -> ^mu+ mu-'

#make collection of functors for Jpsi
variables_jpsi = FunctorCollection({
    'LOKI_P':
    'P',
    'LOKI_PT':
    'PT',
    'LOKI_Muonp_PT':
    'CHILD(PT, 1)',
    'LOKI_Muonm_PT':
    'CHILD(PT, 2)',
    'LOKI_MAXPT':
    'TRACK_MAX_PT',
    'LOKI_N_HIGHPT_TRCKS':
    'NINTREE(ISBASIC & HASTRACK & (PT > 1500*MeV))',
    'THOR_P':
    F.P,
    'THOR_PT':
    F.PT,
    'THOR_MASS':
    F.MASS,
    'DTF_PT':
    F.MAP_INPUT(Functor=F.PT, Relations=DTFRelations),
    'DTF_MASS':
    F.MAP_INPUT(Functor=F.MASS, Relations=DTFRelations),
})

#
# Another way of adding variables.
#
DTF_pv = DTFAlg(Input=dimuons, InputPVs=pvs, MassConstraints=["J/psi(1S)"])
variables_jpsi.update(
    DTF_functors(DTF_pv, functors=[F.PT, F.MASS], head='DTF_PV_'))

#make collection of functors for Muplus
variables_muplus = FunctorCollection({
    'LOKI_P':
    'P',
    'THOR_P':
    F.P,
    'DTF_PT':
    F.MAP_INPUT(Functor=F.PT, Relations=DTFRelations)
})

#associate FunctorCollection to field (branch) name
variables = {}
variables['Jpsi'] = variables_jpsi
variables['MuPlus'] = variables_muplus

#FunTuple: define list of preambles for loki
loki_preamble = ['TRACK_MAX_PT = MAXTREE(ISBASIC & HASTRACK, PT, -1)']

#Configure Funtuple algorithm
tuple_dimuons = Funtuple(
    name="DimuonsTuple",
    tuple_name="DecayTree",
    fields=fields,
    variables=variables,
    loki_preamble=loki_preamble,
    inputs=dimuons)


def main():
    tools = []
    algs = upfront_reconstruction() + [tuple_dimuons]
    return algs, tools
