###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
 Option file for testing the ParticleTaggerAlg algorithm and the related ThOr functors MAP_ARRAY and MAP_RANGE.
 The job runs over a spruced sample and retrieves a set of B0 -> Ds K+ candidates. For each candidate the ParticleTaggerAlg
 looks at the TES location defined via the 'make_long_pions_from_spruce' function and creates a 'one-to-many' relation map
 relating all the available tracks to the B candidate of the events. 
 Then the MAP_ARRAY functor takes in input this relation map and for each entry stores the output of an external functor (i.e F.P, F.PT)
 in a vector via the MAP_RANGE functor.

 This example is meant to be run with
    $ ./run davinci run-mc --inputfiledb Spruce_all_lines_dst Phys/DaVinci/options/DaVinciDB-Example.yaml --user_algorithms DaVinciExamples/python/DaVinciExamples/tupling/option_davinci_tupling_array_taggers:main
"""

import Functors as F
from PyConf.application import make_data_with_FetchDataFromFile
from PyConf.Algorithms import ParticleTaggerAlg, ParticleContainerMerger
from FunTuple import FunctorCollection, FunTuple_Particles as Funtuple
from DaVinci.algorithms import add_filter
from DaVinci.standard_particles import make_long_pions_from_spruce

bd2dsk_line = "SpruceB2OC_BdToDsmK_DsmToHHH_FEST_Line"
bd2dsk_data = make_data_with_FetchDataFromFile(
    f"/Event/Spruce/{bd2dsk_line}/Particles")
pions = make_long_pions_from_spruce()

tagging_container = ParticleContainerMerger(
    InputContainers=[pions]).OutputContainer

tagAlg = ParticleTaggerAlg(
    Input=bd2dsk_data, TaggingContainer=tagging_container, OutputLevel=3)
tagAlg_rels = tagAlg.OutputRelations

#make collection of functors
variables_B = FunctorCollection({
    'THOR_MASS':
    F.MASS,
    "TagTr_P":
    F.MAP_INPUT_ARRAY(Functor=F.MAP_RANGE(Functor=F.P), Relations=tagAlg_rels),
    "TagTr_PT":
    F.MAP_INPUT_ARRAY(
        Functor=F.MAP_RANGE(Functor=F.PT), Relations=tagAlg_rels),
})

#make collection of functors for Muplus
variables_all = FunctorCollection({
    'THOR_P': F.P,
    'THOR_PT': F.PT,
})

fields = {
    'B0': "[B0 -> D_s- K+]CC",
    'Ds': "[B0 -> ^D_s- K+]CC",
    'Kp': "[B0 -> D_s- ^K+]CC",
}

variables = {
    'ALL': variables_all,  #adds variables to all fields
    'B0': variables_B,
}

tuple_B0DsK = Funtuple(
    name="B0DsK_Tuple",
    tuple_name="DecayTree",
    fields=fields,
    variables=variables,
    inputs=bd2dsk_data)

filter_B0DsK = add_filter("HDRFilter_B0DsK",
                          f"HLT_PASS('{bd2dsk_line}Decision')")

from DaVinci import options
options.annsvc_config = 'root://eoslhcb.cern.ch//eos/lhcb/wg/dpa/wp3/tests/spruce_all_lines_realtime.tck.json'
options.histo_file = 'DV-example-tagger-his.root'
options.ntuple_file = 'DV-example-tagger-ntp.root'


def main():
    algs = [filter_B0DsK, tuple_B0DsK]
    return algs, []
