###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Read an HLT2 file and create an ntuple using pre-defined Functor collections.
"""

from PyConf.application import make_data_with_FetchDataFromFile
from FunTuple import FunctorCollection, functorcollections
from FunTuple import FunTuple_Particles as Funtuple
from DaVinci.algorithms import add_filter
from DaVinci import options
from DaVinci.truth_matching import configured_MCTruthAndBkgCatAlg
from DaVinci.algorithms import get_odin, get_decreports

d02kpi_data = make_data_with_FetchDataFromFile(
    "/Event/HLT2/Hlt2CharmD0ToKmPipLine/Particles")

#get configured "MCTruthAndBkgCatAlg" algorithm for HLT2 output
mctruth = configured_MCTruthAndBkgCatAlg(inputs=d02kpi_data)

#get location to odin
odin = get_odin(options)

#get location to hlt2 dec report
dec = get_decreports("Hlt2", options)

# use functorcollections to add variables, please expand when there's new collections :)
collections = [
    functorcollections.Kinematics(),
    functorcollections.MCHierarchy(mctruth),
    functorcollections.MCKinematics(mctruth),
    functorcollections.MCVertexInfo(mctruth)
]

hlt2_line_names = ['Hlt2CharmD0ToKmPipLineDecision']
evt_collections = [
    functorcollections.EventInfo(odin),
    functorcollections.SelectionInfo("Hlt2", dec, hlt2_line_names)
]

assert len(collections) + len(evt_collections) == len(
    functorcollections.
    __all__), "Oh no! Did you forget to add a new collection to this test?"

field_vars = FunctorCollection()
for coll in collections:
    field_vars += coll

evt_vars = FunctorCollection()
for coll in evt_collections:
    evt_vars += coll

fields = {
    "D0": "[D0 -> K- pi+]CC",
    "Kminus": "[D0 -> ^K- pi+]CC",
    "piplus": "[D0 -> K- ^pi+]CC",
}

#For now remove: The 'Hlt2' line decision tuples fine but breaks unit test with an error. (Why?)
#see linked issue here: https://gitlab.cern.ch/lhcb/DaVinci/-/merge_requests/654#note_5320732
print(evt_vars.pop(['Hlt2']))

variables = {
    "D0": field_vars,
    "Kminus": field_vars,
    "piplus": field_vars,
}

my_tuple = Funtuple(
    name="Tuple",
    tuple_name="DecayTree",
    fields=fields,
    variables=variables,
    event_variables=evt_vars,
    inputs=d02kpi_data)


def main():
    my_filter = add_filter("HDRFilter_D0Kpi",
                           "HLT_PASS('Hlt2CharmD0ToKmPipLineDecision')")

    return {"UserAlgs": [my_filter, my_tuple]}, []
