###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Read an HLT2 file and create an ntuple with the new DaVinci configuration.
"""
import Functors as F
from FunTuple import FunctorCollection
from FunTuple import FunTuple_Particles as Funtuple
from PyConf.application import make_data_with_FetchDataFromFile
from DaVinci.reco_objects import make_pvs_for
from DaVinci.algorithms import add_filter
from DaVinci import options
from DaVinci.truth_matching import configured_MCTruthAndBkgCatAlg
from DaVinci.algorithms import get_odin, get_decreports

fields = {
    "D0": "[D0 -> K- pi+]CC",
    "Kminus": "[D0 -> ^K- pi+]CC",
    "piplus": "[D0 -> K- ^pi+]CC",
}

# Creating v2 reconstructed vertices to be used in the following functor
v2_pvs = make_pvs_for(process="Hlt2", data_type="Upgrade")
d0_variables = FunctorCollection({
    "ID": F.PARTICLE_ID,
    "KEY": F.OBJECT_KEY,
    "PT": F.PT,
    "PX": F.PX,
    "PY": F.PY,
    "PZ": F.PZ,
    "ENERGY": F.ENERGY,
    "P": F.P,
    "FOURMOMENTUM": F.FOURMOMENTUM,
    "BPVDIRA": F.BPVDIRA(v2_pvs),
    "BPVFDCHI2": F.BPVFDCHI2(v2_pvs),
    "BPVIPCHI2": F.BPVIPCHI2(v2_pvs)
})

daughter_variables = FunctorCollection({
    "ID": F.PARTICLE_ID,
    "PT": F.PT,
    "PX": F.PX,
    "PY": F.PY,
    "PZ": F.PZ,
    "ENERGY": F.ENERGY,
    "P": F.P,
    "FOURMOMENTUM": F.FOURMOMENTUM,
})

variables = {
    "D0": d0_variables,
    "Kminus": daughter_variables,
    "piplus": daughter_variables
}


def main():
    d02kpi_data = make_data_with_FetchDataFromFile(
        "/Event/HLT2/Hlt2CharmD0ToKmPipLine/Particles")

    my_filter = add_filter("HDRFilter_D0Kpi",
                           "HLT_PASS('Hlt2CharmD0ToKmPipLineDecision')")

    #get configured "MCTruthAndBkgCatAlg" algorithm for HLT2 output
    mctruth = configured_MCTruthAndBkgCatAlg(inputs=d02kpi_data)
    #Add true ID info to each of the fields (branches)
    trueid_bkgcat_info = {
        "TRUEID":
        F.MAP_INPUT(Functor=F.PARTICLE_ID, Relations=mctruth.MCAssocTable),
        "TRUEKEY":
        F.MAP_INPUT(Functor=F.OBJECT_KEY, Relations=mctruth.MCAssocTable),
        "TRUEPT":
        F.MAP_INPUT(Functor=F.PT, Relations=mctruth.MCAssocTable),
        "TRUEPX":
        F.MAP_INPUT(Functor=F.PX, Relations=mctruth.MCAssocTable),
        "TRUEPY":
        F.MAP_INPUT(Functor=F.PY, Relations=mctruth.MCAssocTable),
        "TRUEPZ":
        F.MAP_INPUT(Functor=F.PZ, Relations=mctruth.MCAssocTable),
        "TRUEENERGY":
        F.MAP_INPUT(Functor=F.ENERGY, Relations=mctruth.MCAssocTable),
        "TRUEP":
        F.MAP_INPUT(Functor=F.P, Relations=mctruth.MCAssocTable),
        "TRUEFOURMOMENTUM":
        F.MAP_INPUT(Functor=F.FOURMOMENTUM, Relations=mctruth.MCAssocTable),
        "BKGCAT":
        F.BKGCAT(Relations=mctruth.BkgCatTable)
    }
    for field in variables.keys():
        variables[field] += FunctorCollection(trueid_bkgcat_info)

    #get ODIN and DecReports location
    odin = get_odin(options)
    hlt2_dec = get_decreports("Hlt2", options)

    #Since input is from this line should return 1 for decisions
    hlt2_lines = ['Hlt2CharmD0ToKmPipLineDecision']

    #define event level variables
    evt_variables = FunctorCollection({
        "RUNNUMBER":
        F.RUNNUMBER(odin),
        "EVENTNUMBER":
        F.EVENTNUMBER(odin),
        "Hlt2_TCK":
        F.TCK(hlt2_dec),
        "Sel":
        F.DECISIONS(Lines=hlt2_lines, DecReports=hlt2_dec),
    })
    #For now remove: The 'Hlt2' line decision tuples fine but breaks unit test with an error. (Why?)
    #see linked issue here: https://gitlab.cern.ch/lhcb/DaVinci/-/merge_requests/654#note_5320732
    evt_variables.pop('Sel')

    #define FunTuple instance
    my_tuple = Funtuple(
        name="Tuple",
        tuple_name="DecayTree",
        fields=fields,
        variables=variables,
        event_variables=evt_variables,
        inputs=d02kpi_data)

    return {"UserAlgs": [my_filter, my_tuple]}, []
