###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Read and process a .xgen file with the new DaVinci configuration.

This is run as 
./run davinci run-mc --user_algorithms DaVinciExamples/python/DaVinciExamples/tupling/option_davinci_tupling_from_xgen:main

To test with another xgen file, copy the options above, edit the list of files below (can be local files) and make sure you adapt the database tags. 
Then specify the decay descriptors.
"""
xgenFiles = [
    "root://eoslhcb.cern.ch//eos/lhcb/wg/dpa/wp3/testDV_read_xgen/Gauss-12143001-100ev-20211117.xgen"
]
conddb_tag = "sim-20171127-vc-md100"
dddb_tag = "dddb-20171126"

from FunTuple import FunctorCollection, FunTuple_MCParticles as FuntupleMC
from PyConf.Algorithms import PrintMCTree
from PyConf.dataflow import force_location
import Functors as F

#FunTuple: define fields (branches)
fields = {
    'Bu': "[B+ -> (J/psi(1S) -> mu+ mu-) K+]CC",
    'Jpsi': "[B+ -> ^(J/psi(1S) -> mu+ mu-) K+]CC",
    'Kplus': "[B+ -> (J/psi(1S) -> mu+ mu-) ^K+]CC",
    'muplus': "[B+ -> (J/psi(1S) -> ^mu+ mu-) K+]CC",
    'muminus': "[B+ -> (J/psi(1S) -> mu+ ^mu-) K+]CC",
}

#FunTuple: define variables for the B meson
variables_B = FunctorCollection({
    'ETA': F.ETA,
    'PHI': F.PHI,
    'ORIGIN_VX': F.ORIGIN_VX,
    'ORIGIN_VY': F.ORIGIN_VY,
    'ORIGIN_VZ': F.ORIGIN_VZ,
    'END_VX': F.END_VX,
    'END_VY': F.END_VY,
    'END_VZ': F.END_VZ,
})

#FunTuple: define common variables
variables_all = FunctorCollection({'PT': F.PT, 'P': F.FOURMOMENTUM})

#FunTuple: associate functor collections to field (branch) name
variables = {
    'ALL': variables_all,
    'Bu': variables_B,
}

# Since the "/Event/MC/Particles" location is already declared as output of the UnpackMCParticle algorithm,
# it's not possible to instantiate a new Gaudi__Hive__FetchDataFromFile object with the usual
# 'make_data_with_FetchDataFromFile' method with the same output location.
# It's there sufficient forcing FunTuple to use the same location as input.
bu2jpsik_line = force_location("/Event/MC/Particles")


def main():
    from DaVinci import options

    options.enable_unpack = True
    options.evt_max = -1
    options.histo_file = "DV_histo_xgen.root"
    options.input_raw_format = 0.3
    options.lumi = False
    options.ntuple_file = "DV_tuple_xgen.root"
    options.process = "Hlt2"

    options.input_files = xgenFiles
    options.data_type = "Upgrade"
    options.input_type = "XGEN"
    options.simulation = True
    options.conddb_tag = conddb_tag
    options.dddb_tag = dddb_tag

    printMC = PrintMCTree(
        MCParticles=bu2jpsik_line, ParticleNames=["B+", "B-"], OutputLevel=4)

    tuple_BuJpsiK = FuntupleMC(
        name="BuJpsiK_MCTuple",
        tuple_name="DecayTree",
        fields=fields,
        variables=variables,
        inputs=bu2jpsik_line)

    tools = []
    algs = {"BuJpsiK": [printMC, tuple_BuJpsiK]}
    return algs, tools
