##############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import subprocess, os
import pytest

from pathlib import Path
DIR = Path(__file__).parent.resolve()

from unittest.mock import Mock, patch
input_mock = Mock()


def test_default():
    """
    Check that the simple `davinci` command works.
    """
    result = subprocess.run("davinci", shell=True)
    assert result.returncode == 0


def test_invalid_option():
    """
    Check that an unknown `davinci` option results in an exit code error.
    """
    with pytest.raises(subprocess.CalledProcessError):
        _ = subprocess.run("davinci --unknown", shell=True, check=True)


def test_override_job_option():
    """
    Verify the overriding of a DaVinci option on the fly.
    No need to actually run the job, hence the dry run.
    """
    cmd = 'davinci --export test_override_job_option.opts --dry-run run-mc --override_data_options '\
          ' --process "Turbo" --inputfiledb Upgrade_Bd2KstarMuMu -'
    result = subprocess.run(cmd, shell=True)

    # Just be maniac - the command should work ;-)
    assert result.returncode == 0

    # Inspect the produced .opts file and check for the option value overridden
    with open("test_override_job_option.opts") as f:
        assert any('.process = "Turbo";' in line for line in f.readlines())
    os.remove("test_override_job_option.opts")


def test_override_job_option_bis():
    """
    Verify the overriding of a DaVinci option on the fly,
    no matter the order of the --override_data_options option
    (see related test "test_override_job_option").
    No need to actually run the job, hence the dry run.
    """
    cmd = 'davinci --export test_override_job_option_bis.opts --dry-run run-mc  --process "Turbo" '\
          ' --override_data_options --inputfiledb Upgrade_Bd2KstarMuMu -'
    result = subprocess.run(cmd, shell=True)

    # Just be maniac - the command should work ;-)
    assert result.returncode == 0

    # Inspect the produced .opts file and check for the option value overridden
    with open("test_override_job_option_bis.opts") as f:
        assert any('.process = "Turbo";' in line for line in f.readlines())

    os.remove("test_override_job_option_bis.opts")


def test_run_option_shortcut():
    """
    Verify that the option values passed to the davinci command are accepted 
    also when not using the full name (e.g. --inputfiledb ==> -i).
    """
    cmd = 'davinci --dry-run -o test_run_option_shortcut.opts run-mc -i Upgrade_Bd2KstarMuMu -'
    result = subprocess.run(cmd, shell=True)

    # Just be maniac - the command should work ;-)
    assert result.returncode == 0

    # Inspect the .opts file and check if the input files option is filled correctly.
    with open("test_run_option_shortcut.opts") as f:
        assert not any('.input_files = [];' in line for line in f.readlines())

    os.remove("test_run_option_shortcut.opts")


def test_create_options_templates():
    """
    Verify that the 'create-options-templates' command produces a template 
    for the inputfiledb and joboptfile option files.
    """
    cmd = 'davinci create-options-templates -f test_inputdb_template.yaml test_jobopt_template.yaml'
    result = subprocess.run(cmd, shell=True)

    # Just be maniac - the command should work ;-)
    assert result.returncode == 0

    import yaml
    with open("test_inputdb_template.yaml") as inputdb_file:
        assert yaml.safe_load(inputdb_file)

    with open("test_jobopt_template.yaml") as jobopt_file:
        assert yaml.safe_load(jobopt_file)

    os.remove('test_inputdb_template.yaml')
    os.remove('test_jobopt_template.yaml')


def test_create_template_interactively_changing_defaults(pytestconfig):
    """
    Verify that the interactive method of 'create-joboption-template' command produces
    a correct job option template with the values defined by the user.
    """
    import yaml, os
    from DaVinci.utilities_script import create_jobopt_template
    from DaVinci.options_default import __optsDict__ as dv_defaults

    #Changing some property values from the default for dumping the option file.
    tmp_values = {"annsvc_config": "test_annsvc.tck.json"}

    list_data_qualifiers = [
        "data_type", "input_type", "simulation", "conddb_tag", "dddb_tag",
        "input_files"
    ]

    list_of_args = []
    for name, config in dv_defaults.items():
        if name not in list_data_qualifiers:
            val = tmp_values[name] if name in tmp_values else config["value"]
            list_of_args.append(val)

    input_mock.side_effect = list_of_args

    def test_prompt_values():
        with patch('builtins.input', input_mock):
            create_jobopt_template(DIR / "test_jobopt_template.yaml", True)

    capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')
    capmanager.suspend_global_capture(in_=True)
    test_prompt_values()
    capmanager.resume_global_capture()

    config = []
    with open(DIR / "test_jobopt_template.yaml") as jobopt_file:
        config = yaml.safe_load(jobopt_file)
        assert isinstance(config, dict)
    assert config["annsvc_config"] == "test_annsvc.tck.json"

    os.remove(DIR / 'test_jobopt_template.yaml')
    input_mock.reset_mock(return_value=True, side_effect=True)


def test_create_template_interactively_with_defaults(pytestconfig):
    """
    Verify that the interactive method of 'create-joboption-template' command doesn't produce
    a job option template if all the values are equal to the default ones.
    """
    import os
    from DaVinci.utilities_script import create_jobopt_template
    from DaVinci.options_default import __optsDict__ as dv_defaults

    list_data_qualifiers = [
        "data_type", "input_type", "simulation", "conddb_tag", "dddb_tag",
        "input_files"
    ]

    list_of_args = []
    for name, config in dv_defaults.items():
        if name not in list_data_qualifiers:
            list_of_args.append(config["value"])

    input_mock.side_effect = list_of_args

    def test_prompt_values():
        with patch('builtins.input', input_mock):
            create_jobopt_template(DIR / "test_jobopt_template.yaml", True)

    capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')
    capmanager.suspend_global_capture(in_=True)
    test_prompt_values()
    capmanager.resume_global_capture()

    assert not os.path.exists(DIR / "test_jobopt_template.yaml")
    input_mock.reset_mock(return_value=True, side_effect=True)
