###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
High level configuration tools for DaVinci.
"""
from PyConf.application import ComponentConfig, configure, configure_input
from PyConf.Algorithms import EventAccounting
from GaudiConf import reading
from DaVinci.configOptions import (check_options, set_job_options,
                                   set_properties, set_args_options,
                                   set_input_file_options)
from DaVinci.algorithms import (setup_algorithms, define_fsr_writer,
                                apply_filters_and_unpacking)
from DaVinci.config import davinci_control_flow, prepare_davinci_nodes


def run_davinci_app(fileDB_key="",
                    fileDB_file="",
                    jobOpt_file="",
                    ctx_args=[],
                    simple_job=False,
                    override_data_options=False,
                    prod_conf=False):
    """
    Run the DaVinci application.

    This is the "master function" called

    - by the `davinci` executable in DaVinciSys/scripts.
    - by the `DaVinci` legacy configurable in `DaVinci/compat/DaVinciConf.py`
      for running DaVinci in production via `ProdConf`.
    - directly by the user, typically for simple jobs, when invoking `./run gaudirun.py`.

    Args:
        fileDB_key (str, optional): key in the testfileDB. Defaults to "".
        fileDB_file (str, optional): file with the testfileDB. Defaults to "".
        jobOpt_file (str, optional): file containing the options to be used for running the job.
             Defaults to "".
        ctx_args (iterable, optional): click context containing all the extra options.
             Defaults to [].
        simple_job (bool, optional): flag for running job using only the PyConf configurations.
             Defaults to False.
        override_data_options (bool, optional): boolean for enabling data options override.
             Defaults to False.
        prod_conf (bool, optional): flag for running DaVinci with ProdConf.
             Defaults to False.

    Returns:
        ComponentConfig instance, a dict of configured Configurable instances.
    """
    from DaVinci import options

    # Workaround ConfigurableUser limitation: options.<name> cannot be called if a value is not
    # explicitely assigned and the related property is not set.
    # Initialization in DVAppOptions class seems to be not sufficient.
    set_properties(options)

    if fileDB_key and not prod_conf:
        set_input_file_options(options, fileDB_key, fileDB_file)
    if jobOpt_file:
        set_job_options(options, jobOpt_file, fileDB_key, fileDB_file,
                        override_data_options)
    if ctx_args:
        set_args_options(options, ctx_args, fileDB_key, fileDB_file,
                         override_data_options)

    config = ComponentConfig()

    if simple_job:
        config.update(add_simple_conf(options))
    else:
        config.update(add_davinci_configurables(options))

    if not prod_conf:
        return config


def add_davinci_configurables(options):
    """
    Run the job adding the specific Davinci configurables to the standard PyConf ones.

    Algorithms developed by users are also included.

    Args:
        options (DVAppOptions instance): list of DV options.

    Returns:
        ComponentConfig instance, a dict of configured Gaudi and DaVinci Configurable instances and user algorithms.
    """
    dvMainFlow = {}
    fsrAlgs = {}

    userAlgs, publicTools = setup_algorithms(options)
    # For xgen files we need to unpack only mc particles and vertices
    # Needed until unpacking will become functinal
    unpack_only_mc = options.input_type == "XGEN"

    check_options(options)
    config = configure_input(options)

    if options.annsvc_config:
        config.update(reading.set_hltAnn_svc(options.annsvc_config))
    dvMainFlow.update(
        apply_filters_and_unpacking(options, userAlgs, unpack_only_mc))

    if options.write_fsr:
        if options.simulation:
            fsrAlgs.update({"GenFSR": define_fsr_writer(options)})

        if options.lumi:
            fsrAlgs.update({"Lumi": [EventAccounting(name='EventAccount')]})
            # this should be modified to reflect LumiAlgsConf (configured separately?)

    dvMainNode = davinci_control_flow(options,
                                      prepare_davinci_nodes(dvMainFlow),
                                      prepare_davinci_nodes(fsrAlgs))

    config.update(configure(options, dvMainNode, public_tools=publicTools))
    return config


def add_simple_conf(options):
    """
    Run the job using only the standard PyConf configuration with any user algorithm.

    Args:
        options (DVAppOptions instance): list of DV options.

    Returns:
        ComponentConfig instance, a dict of configured Gaudi Configurable instances and user algorithms.
    """
    userAlgs, publicTools = setup_algorithms(options)

    check_options(options)

    dvNodes = prepare_davinci_nodes(userAlgs)
    top_dv_node = davinci_control_flow(options, dvNodes)

    config = configure_input(options)
    config.update(configure(options, top_dv_node, public_tools=publicTools))

    return config
