###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import os, sys, importlib
from GaudiKernel.ProcessJobOptions import importOptions
from PyConf.Algorithms import (LoKi__HDRFilter as HDRFilter, LoKi__VoidFilter
                               as VoidFilter)
from DaVinci.optionChecker import DVImportError, log_click
from PyConf.application import default_raw_event, make_odin
from PyConf.application import make_data_with_FetchDataFromFile
from PyConf.Algorithms import FilterDecays
from DaVinci.algorithms_pyconf import make_dvalgorithm


def setup_algorithms(options):
    """
    Set DaVinci algorithms.

    Args:
        options: list of DaVinci options.

    Returns:
        Dict of the algorithm instances to be run in the job.
        List of public tool instances to configure.
    """
    opts = options.main_options
    publicTools = []
    dvAlgs = {}

    if not opts == "":
        importOptions(opts)
    else:
        log_click(
            "INFO",
            "No MainOptions specified. DaVinci will import no options file!")

    userAlgName = options.user_algorithms
    if not userAlgName:
        log_click(
            "WARNING",
            "DV option file or main function not defined. No user algorithms will be used."
        )
        # Add a dummy algorithm in order to avoid errors from empty nodes
        from Gaudi.Configuration import ERROR
        from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
        dvAlgs = {"UserAlgorithms": [VoidConsumer(OutputLevel=ERROR)]}
    else:
        userAlgs, publicTools = setup_user_algorithms(userAlgName)
        dvAlgs.update(userAlgs)

    return dvAlgs, publicTools


def set_filter(name, code, dec_reports):
    """
    Adding an HDR filter to FunTuple algorithms checking if the corresponding HLT/Sprucing line fired.

    Args:
        name: filter's name.
        code: filter's code.
        dec_reports: HltDecReportsDecoder containing the configuration for HLT/Sprucing lines.

    Returns:
        Filter with name and code defined by the user.
    """
    if dec_reports:
        algFilter = HDRFilter(
            name=name,
            Code=code,
            Location=dec_reports.OutputHltDecReportsLocation)
    else:
        algFilter = VoidFilter(name=name, Code=code)

    return algFilter


def add_filter(name, code):
    """
    Adding an event pre-filter using a code defined by the user.

    Args:
        name: filter's name.
        code: filter's code.

    Returns:
        Filter with name and code defined by the user.
    """
    import re
    from DaVinci import options
    dec_reports = None
    for source in ["Hlt2", "Spruce"]:
        if re.search("^HLT_PASS.*" + source, code):
            dec_reports = get_hlt_reports(options, source)
            break

    algFilter = set_filter(name, code, dec_reports)
    return algFilter


def apply_filters_and_unpacking(options, algs_dict, unpack_only_mc):
    """
    Adding filter and unpacking algorithms.

    Args:
        options: list of DaVinci options.
        algs_dict: dict of the user algorithms.
        unpack_only_mc: flag to unpack only mc particles and vertices.
        (TO BE REMOVED WHEN THE UNPACKING WILL BECOME FUNCTIONAL)

    Returns:
        Dict where at each node filters and unpacking algorithms are prepended to the initial list of user algorithms.
    """
    alg_filterd_dict = {}
    for name, algs in algs_dict.items():
        algs_list = []
        if options.evt_pre_filters:
            evt_pre_filters = []
            code_filters = options.evt_pre_filters
            for title, code in code_filters.items():
                evt_filter = add_filter(title, code)
                evt_pre_filters.append(evt_filter)
            algs_list += evt_pre_filters

        if options.enable_unpack:
            algs_list += unpack_locations(options, unpack_only_mc)
        algs_list += algs

        alg_filterd_dict[name] = algs_list

    return alg_filterd_dict


def define_fsr_writer(options):
    """
    Define Generator FSR writer.

    Args:
        options: list of DaVinci options.

    Returns:
        List of FSR algorithm instances to be configured.
    """
    from PyConf.Algorithms import GenFSRMerge, RecordStream

    algs = []
    if options.merge_genfsr and options.simulation:
        mergeGenfsr = GenFSRMerge(name="GenFSRMerge")
        algs.append(mergeGenfsr)

    outputLevel = options.output_level
    recStream = RecordStream(
        name="FSROutputStreamDstWriter",
        OutputLevel=outputLevel,
        Output="SVC='Gaudi::RootCnvSvc'")
    algs.append(recStream)

    return algs


def setup_user_algorithms(userAlgPath):
    """
    Set user algorithms and return a PyConf node.

    Args:
        userAlgPath: path to the file containing the algorithms and tools defined by the user.

    Returns:
        Dict of user algorithm instances to be run in the job.
        List of user public tool instances to configure.
    """
    userAlgs = {}
    publicTools = []

    if userAlgPath == "":
        log_click(
            "WARNING",
            "DV option file or main function not defined. No user algorithms will be used."
        )
    else:
        # Identify the module path and the algorithm name from the user defined string
        if '/' in userAlgPath:
            modulePath = userAlgPath.rsplit('/', 1)[0]
            algName = userAlgPath.rsplit('/', 1)[1]
        else:
            modulePath = "./"
            algName = userAlgPath

        # Append module path to sys
        modulePath = os.path.expandvars(modulePath)
        sys.path.append(modulePath)

        if ":" in algName:
            moduleName = algName.rsplit(":", 1)[0]
            funcName = ".%s" % algName.rsplit(":", 1)[1]
        else:
            moduleName = algName
            funcName = ""

        _, ext = os.path.splitext(moduleName)
        if ext == ".py":
            moduleName = moduleName.rsplit('.', 1)[0]

        # Import the module containing the user algorithms
        try:
            module = importlib.import_module(moduleName)  # noqa: F841
        except ModuleNotFoundError:
            raise DVImportError(
                funcName,
                f"Check why the user Python module '{moduleName}' defined in '{modulePath}' can't be imported!"
            ) from None
        else:
            try:
                userAlgs, publicTools = eval("module%s()" % funcName)
            except ImportError:
                raise DVImportError(
                    funcName,
                    f"Runtime error when calling the user algorithm file. User algorithm {funcName} can not be run!"
                )
            else:
                log_click(
                    "INFO", "User algorithm %s%s imported successfully!" %
                    (moduleName, funcName))
                if type(userAlgs) == list:
                    userAlgs = {"UserAlgorithms": userAlgs}

    return userAlgs, publicTools


def unpack_locations(options, unpack_only_mc):
    """
    Configures algorithms for reading HLT2/Spruce output. Location are found using stream as prefix.
    (TO BE REMOVED WHEN THE UNPACKING WILL BECOME FUNCTIONAL)

    Args:
       - options: list of DaVinci options.
       - unpack_only_mc: flag to unpack only mc particles and vertices.

    Returns:
       - List of unpacking algorithms.
    """
    from GaudiConf import reading

    process = options.process
    stream = options.stream

    unpack_raw_event = reading.unpack_rawevent(
        bank_types=['ODIN', 'DstData', 'HltDecReports'],
        process=process,
        stream=stream,
        configurables=False)

    reading_algs = []

    if unpack_only_mc:
        reading_algs += reading.mc_unpackers(
            process=process, filtered_mc=False, configurables=False)
    else:
        reading_algs += [
            unpack_raw_event,
            reading.decoder(
                process=process, stream=stream, configurables=False)
        ]

        if options.simulation:
            reading_algs += reading.mc_unpackers(
                process=process, configurables=False)

        reading_algs += reading.unpackers(process=process, configurables=False)

    return reading_algs


def get_hlt_reports(options, source=''):
    """
    Set the Hlt service and algorithms.

    Args:
       - options: list of DaVinci options.
       - source: source ID required by the user.

    Returns:
       - HltDecReportsDecoder containing the configuration for hlt2 lines.
       - HltDecReportsDecoder containing the configuration for spruced lines.
    """
    from GaudiConf import reading

    process = options.process
    stream = options.stream

    if source == 'Hlt2':
        dec_reports = reading.hlt2_decisions(
            process=process,
            stream=stream,
            output_loc="/Event/%s/DecReports" % source,
            configurables=False)

    elif source == 'Spruce':
        dec_reports = reading.spruce_decisions(
            process=process, stream=stream, configurables=False)

    return dec_reports


def configured_FunTuple(config):
    """
    Function for the FunTuple configuration and instantiation of the related HDR filter.

    Args:
        config (dict): configuration dict containing information for the tuple setting:
          - dict key: name of FunTuple algorithm to be instantiated.
          - tuple: name of the ntuple built by Funtuple.
          - fields: dict containing all the fields (branches) to be stored in Funtuple.
          - variables: dict containing FunctorCollection to be associated to the Funtuple fields (branches).
          - preamble: string for the LoKi preamble
          - location: string containing the particles location to be used as input.
          - filters: list of line decisions for filtering the events.

    Returns:
       - List of filters and tupling algorithms.
    """
    from FunTuple import FunTuple_Particles as Funtuple
    from PyConf.application import make_data_with_FetchDataFromFile

    dictAlgs = {}
    for key in config.keys():
        inputs = make_data_with_FetchDataFromFile(config[key]["location"])
        dictAlgs[key] = []

        i = 0
        for line in config[key]["filters"]:
            filter_name = "Filter_%s" % key
            if len(config[key]["filters"]) > 1:
                filter_name += "_%d" % i
                i = i + 1

            tupleFilter = add_filter(filter_name, line)
            dictAlgs[key].append(tupleFilter)

        funTuple = Funtuple(
            name="Tuple_%s" % key,
            tuple_name=config[key]["tuple"],
            fields=config[key]["fields"],
            variables=config[key]["variables"],
            loki_preamble=config[key]["preamble"],
            inputs=inputs)
        dictAlgs[key].append(funTuple)

    return dictAlgs


def get_odin(dv_options):
    """
    Function to get the LHCb::ODIN location

    Args:
        dv_options: Configured DaVinci.options object
    Returns:
        odin_loc: Location of the LHCb::ODIN
    """
    with default_raw_event.bind(raw_event_format=dv_options.input_raw_format):
        odin_loc = make_odin()

    return odin_loc


def get_decreports(sel_stage, dv_options):
    """
    Function to get the LHCb::DecReports for HLT1 or Hlt2 or Spruce

    Args:
        sel_state (str): Selction stage can be "Hlt1" or "Hlt2" or "Spruce" (There does not seem to be a decoder for Hlt1 currently)
        dv_options: Configured DaVinci.options object
    Returns:
        dec_loc: Location of the LHCb::DecReports for HLT1 or Hlt2 or Spruce
    """
    dec_loc = get_hlt_reports(
        dv_options, source=sel_stage).OutputHltDecReportsLocation
    return dec_loc


def apply_algorithm(list_particles, algorithm, **kwargs):
    """
    Function that applies an algorithm, accepting a Particle::Range as input (e.g FilterDecays),
    to particles from Hlt2 or Spruce sample defined in the list of TES locations gived as argument.
    The extra arguments to algorithm can be passed via kwargs.

    Args:
        list_particles (list(str)): List of Transient event store (TES) location of Particles in Hlt2 or Spruce sample.
        algorithm (PyConf.Algorithms): Algorithm to be applied.
        kwargs : Extra arguments to the passed algorithm.
    Returns:
        Ouput TES location of the particles from the algorithm
    """
    dv_algorithm = make_dvalgorithm(algorithm)
    return dv_algorithm(particles=list_particles, **kwargs).particles


def filter_on(location, decay_descriptor=None, bank_type=None):
    """
    Function to get particles from Hlt2 or Spruce sample.
    A FilterDecays is applied before returning the requested particles if a decay descriptor
    is specified as extra argument.

    Args:
        location (str): Transient event store (TES) location of Particles in Hlt2 or Spruce sample
        decay_descriptor (str, optional): Decay descriptor that will be used to select decays in the event
        bank_type (str, optional): Raw bank type
    Returns:
        data: TES location of the particles that are loaded from the input samples
    """
    data = make_data_with_FetchDataFromFile(location, bank_type=bank_type)
    if decay_descriptor:
        data = apply_algorithm([data], FilterDecays, Code=decay_descriptor)
    return data
