###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Selection and combiner wrappers.

Algorithms that inherit from DVCommonBase, like FilterDesktop and
CombineParticles, are not functional and do not expose input/output
DataHandles. They also do some funky internal location mangling to save
additional objects next to the Particle objects they create. The wrappers here
try to work around these traits to make the algorithms behave like any other
functional algorithms.
"""
from PyConf.Algorithms import (CombineParticles, FilterDesktop,
                               DaVinci__N3BodyDecays as N3BodyDecays,
                               DaVinci__N4BodyDecays as N4BodyDecays)

__all__ = [
    #'EmptyFilter', 'ParticleFilter', 'ParticleCombiner',
    'ParticleFilter',
    'ParticleCombiner',
    'ParticleFilterWithPVs',
    'ParticleCombinerWithPVs',
    'require_all',
    'N3BodyCombiner',
    'N3BodyCombinerWithPVs',
    'N4BodyCombiner',
    'N4BodyCombinerWithPVs',
    'NeutralParticleCombiner',
    'NeutralParticleCombinerWithPVs'
]


def require_all(*cuts):
    """Return a cut string requiring all arguments.

    Example:

        >>> require_all('PT > {pt_min}', 'DLLK < {dllk_max}')
        '(PT > {pt_min}) & (DLLK < {dllk_max})'
    """
    cuts = ['({})'.format(c) for c in cuts]
    return ' & '.join(cuts)


def _dvalgorithm_inputs(particles, pvs=None):
    """
    Return a dict suitable for a DVAlgorithm input transform.
    
    Args:
        particles (list): list of particle containers used as input to the DV algorithm
        pvs (optional): primary vertices container

    Returns:
        Dict containing both particles and primary vertices containers
    """
    # ExtraInputs is added by the data handle mixin, so we bundle all inputs
    # there to make them available to the scheduler
    d = {'Inputs': particles, 'ExtraInputs': particles}
    if pvs:
        d['InputPrimaryVertices'] = pvs
    return d


def _dvalgorithm_outputs(particles):
    """
    Return a dict suitable for a DVAlgorithm output transform.

    Args:
        particles: output particles container created by the DV algorithm.

    Returns:
        Dict containing the information on the output container.
    """
    # ExtraOutputs is added by the data handle mixin, so we can add the output
    # there to make it available to the scheduler
    # Could add, for example, output P2PV relations or refitted PVs here as
    # well
    d = {'Output': particles, 'ExtraOutputs': [particles]}
    return d


def make_dvalgorithm(algorithm):
    """
    Function creating a wrapper for the specified algorithm.
    
    Args:
        algorithm: PyConf.Algorithm instance of the algorithm of interest.

    Returns:
        Wrapped instance of the algorithm specified in input.
    """

    def wrapped(**kwargs):
        return algorithm(
            input_transform=_dvalgorithm_inputs,
            output_transform=_dvalgorithm_outputs,
            WriteP2PVRelations=False,
            ModifyLocations=False,
            **kwargs)

    return wrapped


combiner = make_dvalgorithm(CombineParticles)


def ParticleFilter(particles, **kwargs):
    """
    Return a filter algorithm that takes `particles` as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        kwargs: additional keyword arguments are forwarded to FilterDesktop.

    Returns:
        Container of the particles surviving the FilterDesktop.
    """
    filter_desktop = make_dvalgorithm(FilterDesktop)
    particles = particles if isinstance(particles, list) else [particles]
    inputs = {'particles': particles}
    # Assert kwargs doesn't containt other elements named particles to avoid
    # conflicts in the input particles definition
    assert set(inputs).intersection(kwargs) == set()
    kwargs = dict(list(inputs.items()) + list(kwargs.items()))

    return filter_desktop(**kwargs).particles


def ParticleFilterWithPVs(particles, pvs, **kwargs):
    """
    Return a filter algorithm that takes `particles` and `pvs` as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        pvs: container of the primary vertices.
        kwargs: additional keyword arguments are forwarded to FilterDesktop.

    Returns:
        Container of the particles surviving the FilterDesktop.
    """
    return ParticleFilter(particles=particles, pvs=pvs, **kwargs)


def ParticleCombiner(particles, combiner=combiner, **kwargs):
    """
    Return a combiner algorithm that takes `particles` as inputs.
    Args:
        particles (list): list of particle containers to be filtered.
        combiner: algorithm for combining input particles.
        kwargs: additional keyword arguments are forwarded to CombineParticles.

    Returns:
        Container of the particles generated by the CombineParticles algorithm.
    """
    particles = particles if isinstance(particles, list) else [particles]
    inputs = {'particles': particles}
    # We need to merge dicts, we make sure we don't have overlapping keys (the
    # caller really shouldn't specify Particles keys anyway)
    assert set(inputs).intersection(kwargs) == set()
    kwargs = dict(list(inputs.items()) + list(kwargs.items()))

    return combiner(**kwargs).particles


def N3BodyCombiner(particles, **kwargs):
    """
    Return a N3BodyDecays combiner algorithm that takes particles as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        kwargs: additional keyword arguments are forwarded to ParticleCombiner.

    Returns:
        Container of the particles generated by the N3BodyDecays combiner.
    """
    threebodycombiner = make_dvalgorithm(N3BodyDecays)
    return ParticleCombiner(particles, combiner=threebodycombiner, **kwargs)


def N4BodyCombiner(particles, **kwargs):
    """
    Return a N4BodyDecays combiner algorithm that takes particles as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        kwargs: additional keyword arguments are forwarded to ParticleCombiner.

    Returns:
        Container of the particles generated by the N4BodyDecays combiner.
    """
    fourbodycombiner = make_dvalgorithm(N4BodyDecays)
    return ParticleCombiner(particles, combiner=fourbodycombiner, **kwargs)


def ParticleCombinerWithPVs(particles, pvs, **kwargs):
    """
    Return a combiner algorithm that takes `particles` and `pvs` as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        pvs: container of the primary vertices.
        kwargs: additional keyword arguments are forwarded to CombineParticles.

    Returns:
        Container of the particles generated by the CombineParticles algorithm.
    """
    return ParticleCombiner(particles=particles, pvs=pvs, **kwargs)


def N3BodyCombinerWithPVs(particles, pvs, **kwargs):
    """
    Return a combiner algorithm that takes `particles` and `pvs` as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        pvs: container of the primary vertices.
        kwargs: additional keyword arguments are forwarded to N3BodyCombiner.

    Returns:
        Instance of N3BodyCombiner
    """
    ## TODO:  eliminate duplication of code with ParticleCombinerWithPVs
    return N3BodyCombiner(particles=particles, pvs=pvs, **kwargs)


def N4BodyCombinerWithPVs(particles, pvs, **kwargs):
    """
    Return a combiner algorithm that takes `particles` and `pvs` as inputs.

    Args:
        particles (list): list of particle containers to be filtered.
        pvs: container of the primary vertices.
        kwargs: additional keyword arguments are forwarded to N4BodyCombiner.

    Returns:
        Instance of N4BodyCombiner.
    """
    ## TODO:  eliminate duplication of code with ParticleCombinerWithPVs
    return N4BodyCombiner(particles=particles, pvs=pvs, **kwargs)


def NeutralParticleCombinerWithPVs(particles, pvs, **kwargs):
    """
    Return a combiner algorithm that takes `particles` and `pvs` as inputs.
    No vertex fit is performed, just momentum addition

    Args:
        particles (list): list of particle containers to be filtered.
        pvs: container of the primary vertices.
        kwargs: additional keyword arguments are forwarded to NeutralParticleCombiner.

    Returns:
        Instance of NeutralParticleCombiner.
    """
    return NeutralParticleCombiner(particles=particles, pvs=pvs, **kwargs)


def NeutralParticleCombiner(particles, **kwargs):
    """
    Return a combiner algorithm that takes `particles` as input.
    No vertex fit is performed, just momentum addition
    
    Args:
        particles (list): list of particle containers to be filtered.
        kwargs: additional keyword arguments are forwarded to CombineParticles.

    Returns:
        Container of the particles generated by the CombineParticles algorithm.
    """
    return ParticleCombiner(
        particles=particles, ParticleCombiners={"": "ParticleAdder"}, **kwargs)
