###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Definitions of "common particles" very similar to those of Runs 1 & 2.
"""

from PyConf.tonic import configurable
from PyConf.Algorithms import FunctionalParticleMaker
from PyConf.Algorithms import LHCb__Phys__ParticleMakers__PhotonMaker as PhotonMaker

from .reco_objects import make_charged_protoparticles as _make_charged_protoparticles
from .reco_objects import make_neutral_protoparticles as _make_neutral_protoparticles
from .reco_objects import make_pvs as _make_pvs

from .filters_selectors import default_particle_cuts, default_track_cuts
from .filters_selectors import get_long_track_selector, get_down_track_selector
from .filters_selectors import all_protoparticle_filter as standard_protoparticle_filter

from .algorithms_pyconf import ParticleFilterWithPVs, ParticleCombinerWithPVs

#########
# Helpers
#########


@configurable
def _make_particles(species,
                    make_protoparticles=_make_charged_protoparticles,
                    get_track_selector=get_long_track_selector,
                    make_protoparticle_filter=standard_protoparticle_filter):
    """
    Helper configurable to create `LHCb::Particle`s from `LHCb::ProtoParticle`s.

    Args:
        species (str): Particle species hypothesis accepted by
            `FunctionalParticleMaker`, i.e. one of the strings
            "pion", "kaon", "muon", "electron", "proton".
    """
    particles = FunctionalParticleMaker(
        ParticleID=species,
        InputProtoParticles=make_protoparticles(),
        TrackSelector=get_track_selector(),
        ProtoParticleFilter=make_protoparticle_filter()).Particles
    return particles


@configurable
def make_photons(make_neutral_protoparticles=_make_neutral_protoparticles,
                 pvs=_make_pvs,
                 **kwargs):
    """
    Configurable to create photon `LHCb::Particle`s from `LHCb::ProtoParticle`s.
    """
    particles = PhotonMaker(
        InputProtoParticles=make_neutral_protoparticles(),
        InputPrimaryVertices=pvs(),
        **kwargs).Particles
    return particles


def _make_std_loose_particles(particles, pvs, name):
    return ParticleFilterWithPVs(
        particles, pvs, name=name, Code=default_particle_cuts())


#######################
# Bacic particle makers
#######################


def make_long_pions():
    return _make_particles(species="pion")


def make_long_kaons():
    return _make_particles(species="kaon")


def make_long_protons():
    return _make_particles(species="proton")


def make_long_muons():
    return _make_particles(species="muon")


def make_long_electrons_no_brem():
    return _make_particles(species="electron")


def make_down_pions():
    return _make_particles(
        species="pion", get_track_selector=get_down_track_selector)


def make_down_kaons():
    return _make_particles(
        species="kaon", get_track_selector=get_down_track_selector)


def make_down_protons():
    return _make_particles(
        species="proton", get_track_selector=get_down_track_selector)


#################################
# Particle makers with loose cuts
#################################


@configurable
def make_std_loose_pions():
    with get_long_track_selector.bind(
            Code=default_track_cuts()), standard_protoparticle_filter.bind(
                Code='PP_HASRICH'):
        return _make_std_loose_particles(
            make_long_pions(), _make_pvs(), name='StdLoosePions')


@configurable
def make_std_loose_kaons():
    with get_long_track_selector.bind(
            Code=default_track_cuts()), standard_protoparticle_filter.bind(
                Code='PP_HASRICH'):
        return _make_std_loose_particles(
            make_long_kaons(), _make_pvs(), name='StdLooseKaons')


@configurable
def make_std_loose_protons():
    with get_long_track_selector.bind(
            Code=default_track_cuts()), standard_protoparticle_filter.bind(
                Code='PP_HASRICH'):
        return _make_std_loose_particles(
            make_long_protons(), _make_pvs(), name='StdLooseProtons')


def make_std_loose_muons():
    #with get_long_track_selector.bind(Code=default_track_cuts()):
    return _make_std_loose_particles(
        make_long_muons(), _make_pvs(), name='StdLooseMuons')


@configurable
def make_std_loose_jpsi2mumu():
    muons = make_std_loose_muons()
    descriptors = ["J/psi(1S) -> mu+ mu-"]
    daughters_code = {"mu+": "ALL", "mu-": "ALL"}
    combination_code = "(ADAMASS('J/psi(1S)') < 100.*MeV) & (ADOCACHI2CUT(30,''))"
    vertex_code = "(CHI2VX < 25.)"

    return ParticleCombinerWithPVs(
        name="StdLooseJpsi2MuMu",
        particles=muons,
        pvs=_make_pvs(),
        DecayDescriptors=descriptors,
        DaughtersCuts=daughters_code,
        CombinationCut=combination_code,
        MotherCut=vertex_code)


@configurable
def make_std_loose_d2kk():
    kaons = make_std_loose_kaons()
    descriptors = ["D0 -> K+ K-"]
    daughters_code = {"K+": "ALL", "K-": "ALL"}
    combination_code = "(ADAMASS('D0') < 100.*MeV) & (ADOCACHI2CUT(30,''))"
    vertex_code = "(CHI2VX < 25.)"

    return ParticleCombinerWithPVs(
        name="StdLooseD02KK",
        particles=kaons,
        pvs=_make_pvs(),
        DecayDescriptors=descriptors,
        DaughtersCuts=daughters_code,
        CombinationCut=combination_code,
        MotherCut=vertex_code)
