###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
DaVinci configuration for ProdConf.
"""
import re
from Gaudi.Configuration import appendPostConfigAction
from Configurables import (LHCbApp, EventSelector, DDDBConf, CondDB,
                           LHCbConfigurableUser, HistogramPersistencySvc)
from DaVinci.Configuration import run_davinci_app
from DaVinci.configOptions import set_option_value
from DaVinci import options

from AnalysisPython.Logger import getLogger
log = getLogger('DaVinci')


class DaVinciConfigurationError(Exception):
    """Exception type raised when DaVinci is not configured consistently."""
    pass


class DaVinci(LHCbConfigurableUser):

    __used_configurables__ = [
        LHCbApp,
        DDDBConf,
        CondDB,
    ]

    __slots__ = {
        "CondDBtag": "",
        "DDDBtag": "",
        "EvtMax": -1,
        "HistogramFile": "",
        "OutputFile": "",
    }

    _propertyDocDct = {
        "CondDBtag":
        """ Tag for the CondDB. """,
        "DDDBtag":
        """ Data type, can be ["Upgrade"] Forwarded to PhysConf, AnalysisConf, DstConf and LHCbApp. """,
        "EvtMax":
        """ Number of events to analyse. Default = -1 to run over all events. """,
        "HistogramFile":
        """Name of output histogram file. Default = ''.""",
        "OutputFile":
        """ Name of output file. Default = ''.""",
    }

    _optNameMap = {
        "CondDBtag": "conddb_tag",
        "DDDBtag": "dddb_tag",
        "EvtMax": "evt_max",
        "HistogramFile": "histo_file",
        "Input": "input_files",
        "OutputFile": "ntuple_file",
    }

    def undressInputFileName(self, selector):
        m = re.match(r"DATAFILE='([^']+)'", selector)
        return m.group(1)

    def remove_duplicates(self):
        """Remove duplicates introduced by using both PyConf and LHCbApp"""
        from Gaudi.Configuration import allConfigurables
        for name, prop in (("ApplicationMgr",
                            "ExtSvc"), ("EventPersistencySvc", "CnvServices"),
                           ("FileRecordPersistencySvc", "CnvServices")):
            try:
                c = allConfigurables[name]
                # we have a mixture of configurables and strings,
                # so we count as duplicates the entries with the same name
                new_items = {}
                for element in getattr(c, prop):
                    el_name = element.name() if hasattr(
                        element, "name") else element.rsplit("/", 1)[-1]
                    new_items.setdefault(el_name, element)
                setattr(c, prop, list(new_items.values()))
            except (AttributeError, KeyError):
                pass

    def retrieve_options(self):
        """
        Retrieve option values from LHCbApp().
        """
        for name in self._optNameMap:
            if name in ["CondDBtag", "DDDBtag", "EvtMax"]:
                value = LHCbApp().getProp(name)
                self.setProp(name, value)
            elif name == "HistogramFile":
                if HistogramPersistencySvc().isPropertySet("OutputFile"):
                    value = HistogramPersistencySvc().OutputFile
                    self.setProp(name, value)

    def propagate_options(self, options):
        """
        Propagate DaVinci properties from __slots__ to DaVinci.options.
        """
        for name in self._optNameMap:
            if name == "Input":
                inputs = [
                    self.undressInputFileName(s) for s in EventSelector().Input
                ]
                set_option_value(options, self._optNameMap[name], inputs)
            else:
                if self.getProp(name):
                    set_option_value(options, self._optNameMap[name],
                                     self.getProp(name))

    def run_davinci_conf(self):
        """
        Run the DaVinci configuration.
        """
        run_davinci_app(prod_conf=True)

    def __apply_configuration__(self):
        """
        DaVinci configuration.
        """
        log.info("Applying DaVinci configuration")

        self.retrieve_options()
        log.info(self)

        self.propagate_options(options)
        self.run_davinci_conf()

        appendPostConfigAction(self.remove_duplicates)
