###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DaVinci configured using PyConf components.
"""

from __future__ import absolute_import
import logging
from collections import namedtuple, OrderedDict
from PyConf.application import configure_input, configure
from PyConf.control_flow import CompositeNode, NodeLogic

from DaVinci.application import DVAppOptions

log = logging.getLogger(__name__)

options = DVAppOptions(_enabled=False)


class DVNode(namedtuple('DVNode', ['node', 'extra_outputs'])):  # noqa
    """Immutable object fully qualifying a DaVinci node.
    Copied from the `HltLine` class without the prescaler argument.

    Attributes:
        node (CompositeNode): the control flow node of the line

    """
    __slots__ = ()  # do not add __dict__ (preserve immutability)

    def __new__(cls, name, algs, extra_outputs=None):
        """Initialize a DaVinci node from name and a set of algorithms.

        Creates a control flow `CompositeNode` with the given `algs`
        combined with `LAZY_AND` logic.

        Args:
            name (str): name of the line
            algs: iterable of algorithms
            extra_outputs (iterable of 2-tuple): List of (name, DataHandle) pairs.
        """
        node = CompositeNode(
            name,
            tuple(algs),
            combine_logic=NodeLogic.LAZY_AND,
            force_order=True)
        if extra_outputs is None:
            extra_outputs = []
        return super(DVNode, cls).__new__(cls, node, frozenset(extra_outputs))

    @property
    def name(self):
        """DVNode (CompositeNode instance) name."""
        return self.node.name


def davinci_control_flow(options, user_analysis_nodes=[], fsr_nodes=[]):
    """
    DaVinci control flow is split in a few sections as described in DaVinci/issue#2 (then simplified)

       DaVinci (LAZY_AND)
       *-- LuminosityNode (NONLAZY_OR)
       |   *-- EventAccounting/EventAccount
       *-- UserAnalysisNode (NONLAZY_OR)
           +-- DVUser1Node (LAZY_AND)
           |   *-- PVFilter
           |   *-- ParticlesFilter
           |   *-- Candidate1Combiner
           |   *-- AlgorithmsForTuple1
           |   *-- Tuple1
           +-- DVUser2Node (LAZY_AND)
               *-- PVFilter
               *-- ParticlesFilter
               *-- Candidate2Combiner
               *-- AlgorithmsForTuple2
               *-- Tuple2

    The main parts are
        . LuminosityNode
        . UserAnalysisNode
    who are in AND among each other and can accomodate nodes (defined with the class DVNode)
    in OR among themselves.

    To prepare the control flow there are a few options:
        1. take a dictionary as input where all the nodes are listed in their categories
        2. take a various lists as input each related to a specific category of nodes
    This function is then used to fill the control flow
    """
    options.finalize()

    dv_top_children = []

    ordered_nodes = OrderedDict()
    ordered_nodes['FileSummaryRecords'] = fsr_nodes
    ordered_nodes['UserAnalysis'] = user_analysis_nodes

    for k, v in ordered_nodes.items():
        if len(v):
            cnode = CompositeNode(
                k,
                combine_logic=NodeLogic.NONLAZY_OR,
                children=[dv_node.node for dv_node in v],
                force_order=False)
            dv_top_children += [cnode]

    return CompositeNode(
        'DaVinci',
        combine_logic=NodeLogic.LAZY_AND,
        children=dv_top_children,
        force_order=True)


def prepare_davinci_nodes(user_algs):
    """
    This helper function takes as input a dictionary of user algorithms in the form
    {
        'DVNode1' : [<list of algs>],
        'DVNode2' : [<list of algs>],
    }
    and creates the node to send to `davinci_control_flow`.
    """
    dv_nodes = []
    for k, algs in user_algs.items():
        if not type(algs) == list:
            raise TypeError("Dict values should all be lists of algorithms!")
        dv_nodes += [DVNode(k, algs)]
    return dv_nodes


def run_davinci(options, user_algs, public_tools=[]):
    """
    DaVinci application control flow.

    Args:
        options (ApplicationOptions): holder of application options.
        user_algs (dict): list of user algorithms 'catalogued' with the dict key,
            e.g. {'MyAlgs': [<list of algs>]}.
        public_tools (list): list of public `Tool` instances to configure.
    """
    config = configure_input(options)

    dv_nodes = prepare_davinci_nodes(user_algs)

    top_dv_node = davinci_control_flow(options, dv_nodes)

    config.update(configure(options, top_dv_node, public_tools=public_tools))

    return config
