###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Set and retrieve the value of a specific option property.
"""

import os
from DaVinci.optionChecker import DVOptionError, DVRuntimeError
from DaVinci.optionChecker import option_checker, log_click


def set_option_value(options, name, value):
    """
    Set option value for a given name.

     Args:
         options: list of DaVinci options.
         name: name of the option of interest.
         value: value to be set.
    """
    # Properties value passed by command line are always interpreted as string
    # Numerical values are converted automatically but this is not true for boolean
    # So a conversion of the values from string to boolean is applied when needed
    if isinstance(options.getProp(name), bool) and isinstance(value, str):
        if value in {"True", "False"}:
            value = (value == "True")
        else:
            raise DVOptionError(
                name,
                f"'{name}' is boolean and can accept only 'True' or 'False' as value"
            )

    options.setProp(name, value)


def set_input_file_options(options, fileDB_key, fileDB_file):
    """
    Set the dataset properties required by the user.

    This helper function is similar to what `set_input_and_conds_from_testfiledb` in
    `https://gitlab.cern.ch/lhcb/LHCb/-/blob/master/PyConf/python/PyConf/application.py`
    does, but the latter acts on a TestFileDB entry.

    Args:
        options: list of DaVinci options.
        fileDB_key: key in the testfileDB.
        fileDB_file: file containing the testfileDB.
    """
    if fileDB_file == "TestFileDB":
        options.set_input_and_conds_from_testfiledb(fileDB_key)
    else:
        import yaml
        with open(os.path.expandvars(fileDB_file)) as fileDB:
            test_file_db = yaml.safe_load(fileDB)

        # File index for running job only on a specific file in the testfileDB list.
        # File number specified with ":" (e.g. key:2). Dafault -1 for running over all the files.
        idxFile = -1
        if ":" in fileDB_key:
            idxFile = int(fileDB_key.split(":")[1])
            fileDB_key = fileDB_key.split(":")[0]

        # Given a specific key, loop over the elements of the testfileDB
        for item, obj in test_file_db[fileDB_key].items():
            # Set all the file quelifiers:
            if item == 'qualifiers':
                for prop, value in obj.items():
                    set_option_value(options, prop, value)
            # set the input files:
            elif item == 'filenames':
                # if file index is not -1, select only the corresponding file from the list in testfileDB.
                if idxFile > -1:
                    # Check that the index element exists
                    if idxFile < len(obj):
                        set_option_value(options, "input_files",
                                         [obj[idxFile]])
                    else:
                        raise ValueError(
                            'Index file exceeds the number of available files related to the given key!'
                        )
                else:
                    set_option_value(options, "input_files", obj)


def set_job_options(options, jobOptFile, fileDB_key, fileDB_file,
                    override_data_options):
    """
    Set the job properties required by the user. The method checks if there
    are options related to the input files selected with key and file of the
    testfileDB, if yes it checks if the option is settable.

    Args:
        options: list of DaVinci options.
        jobOptFile: file containing the job options chosen by the user.
        fileDB_key: key in the testfileDB.
        fileDB_file: file containing the testfileDB.
        override_data_options: boolean for enabling data options override.
    """
    dataOptions = list_data_options(fileDB_key.split(":")[0], fileDB_file)

    with open(os.path.expandvars(jobOptFile)) as config_file:
        _, ext = os.path.splitext(jobOptFile)
        if ext in (".yaml", ".yml", ".json"):
            import yaml
            config = yaml.safe_load(config_file)
        elif ext == '.py':
            import ast
            config = ast.literal_eval(config_file.read())
        else:
            raise ValueError(
                'JobOption file extension not known! Please use only a .py, .json or .yaml (.yml) file!'
            )

        for key, value in config.items():
            if is_option_settable(options, key, dataOptions,
                                  override_data_options):
                set_option_value(options, key, value)


def set_args_options(options, ctx_args, fileDB_key, fileDB_file,
                     override_data_options):
    """
    Set the extra arguments required by the user.

    Args:
        options: list of DaVinci options.
        ctx_args: click context.
        fileDB_key: key in the testfileDB.
        fileDB_file: file containing the testfileDB.
        override_data_options: boolean for enabling data options override.
    """
    dataOptions = list_data_options(fileDB_key.split(":")[0], fileDB_file)

    # Loop over the properties and set their values
    for i in range(0, len(ctx_args), 2):
        key = ctx_args[i][2:]
        value = ctx_args[i + 1]

        if is_option_settable(options, key, dataOptions,
                              override_data_options):
            set_option_value(options, key, value)


def is_option_settable(options, name, dataOptions, override_data_options):
    """
    Check if the option identified by name is settable.
    This is true for all the job options while for data options
    the 'override_data_options' flag is required.
    Indeed, by default all the data options are already set
    using the information found in the testfileDB.

    Args:
        options: list of DaVinci options.
        name: name of the option to be checked.
        dataOptions: list of options related to the input data.
        override_data_options: boolean for enabling data options override.
    Raises:
        DVRuntimeError: if 'override_data_options' is set and input 'name' is a data option.
    """
    if name in dataOptions:
        if override_data_options:
            log_click(
                "INFO",
                "New value found for the option %s in the job option file. "\
                "'override_data_options' is active so the default value will be overridden."
                % name)
            return True
        else:
            raise DVRuntimeError(
                name, "override_data_options",
                "Default value for option %s is taken from DaVinciDB.\nA new value is found in the job "\
                "option file but 'override_data_options' is not active so the default value can't be overridden!"
                % name)
    else:
        return True


def list_data_options(fileDB_key, fileDB_file):
    """
    Get the list of the properties that are set automatically given a dataset.

    Args:
        fileDB_key: key in the testfileDB.
        fileDB_file: file containing the testfileDB.
    """
    import yaml

    optionList = []
    qualifiers = []
    if fileDB_key:
        if fileDB_file == 'TestFileDB':
            from PRConfig.TestFileDB import test_file_db
            qualifiers = test_file_db[fileDB_key].qualifiers
        else:
            with open(os.path.expandvars(fileDB_file)) as fileDB:
                test_file_db = yaml.safe_load(fileDB)
                qualifiers = test_file_db[fileDB_key]["qualifiers"]

        for prop, value in qualifiers.items():
            optionList.append(prop)

    return optionList


def check_options(options):
    """
    Check the options. Applies changes if needed.

    Args:
        options: list of DaVinci options.
    """
    dataType = options.data_type
    option_checker("data_type", dataType)

    inputType = (options.input_type).upper()
    option_checker("input_type", inputType)
    if inputType != "MDF":
        set_option_value(options, "input_type", "ROOT")

    flagMC = options.simulation
    if not flagMC:
        if options.merge_genfsr:
            raise DVRuntimeError(
                "merge_genfsr", "simulation",
                "GenFSR are not available in real data so merge_genfsr cannot be run! Set merge_genfsr = False."
            )
    else:
        if options.lumi:
            raise DVRuntimeError(
                "lumi", "simulation",
                "lumi not valid for simulation! Set lumi = False.")

        ## for simulation, it is very important to specify proper DB-tags:
        if not options.dddb_tag:
            raise DVOptionError(
                "dddb_tag", "'dddb_tag' is not specified for simulated data!")
        if not options.conddb_tag:
            raise DVOptionError(
                "conddb_tag",
                "'conddb_tag' is not specified for simulated data!")


def set_properties(options):
    for name, default in options.getDefaultProperties().items():
        if not options.isPropertySet(name):
            options.setProp(name, default)
