###############################################################################
# (c) Copyright 2019-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Helper module with functions to load reco./MC data and linker tables from files,
and set up reconstruction and simulation unpackers.

There are two things we have to deal with:

1. Loading the data from the file in to the TES, done by
   `Gaudi::Hive::FetchDataFromFile`.
2. Unpacking and preparing packed containers, if the 'reconstruction' is
   defined as the objects already present in the file.

In most LHCb applications step 2 is done behind the scenes:
the `DataOnDemandSvc` is configured in `LHCb/GaudiConf/DstConf.py`
to unpack containers when they are requested.
It also configures adding RICH, MUON, and combined PID information to ProtoParticles
when the unpacking takes place. This module effectively does all these steps
explicitly because the `PyConf` framework does not rely (by construction!)
on the somewhat subtle and obscure `DataOnDemandSvc`.

The interesting "user-facing" exports of this module are
`{reco,mc}_unpackers()`, which return a dict from unpacked object name to
a `PyConf.Algorithm` instance that produces a container of those objects.

 The locations are defined under `DaVinci.locations`.

.. note::
    1) The functions defined in this module rely on data paths used in Runs 1 & 2,
    and may need a revision once the Run 3 event model is finalised
    and the definition of what gets persisted gets formalised.
    2) Code very heavily relies on its Moore equivalent. Thank you, RTA team.
"""
from __future__ import absolute_import, division, print_function

import collections

from Gaudi.Configuration import ERROR

from Configurables import (UnpackCaloHypo, UnpackProtoParticle,
                           UnpackRecVertex, UnpackTrack, UnpackMCParticle,
                           UnpackMCVertex)
from Configurables import (
    MuonPIDUnpacker, RichPIDUnpacker, MCVPHitUnpacker as UnpackMCVPHit,
    MCUTHitUnpacker as UnpackMCUTHit, MCFTHitUnpacker as UnpackMCFTHit,
    MCRichHitUnpacker as UnpackMCRichHit, MCEcalHitUnpacker as UnpackMCEcalHit,
    MCHcalHitUnpacker as UnpackMCHcalHit, MCMuonHitUnpacker as UnpackMCMuonHit,
    MCRichDigitSummaryUnpacker as RichSumUnPack)

from PyConf.Tools import (ChargedProtoParticleAddRichInfo,
                          ChargedProtoParticleAddMuonInfo,
                          ChargedProtoParticleAddCombineDLLs)

from PyConf.components import Algorithm, force_location
from PyConf.application import make_data_with_FetchDataFromFile

from .locations import (LocationsPackedReco, LocationsUnpackedReco)
from .locations import (LocationsPackedSim, LocationsUnpackedSim)
from .locations import (LocationsBooleMCParticleLinkers,
                        LocationsBooleMCHitsLinkers, LocationsBrunelMCLinkers,
                        LocationMCTrackInfo)
from .locations import enums_as_dict


def reco_unpackers():
    """
    Return a {object name: `PyConf.Algorithm` instance}  `OrderedDict`
    effectively mapping unpacked reconstruction object names to their respective
    unpacked data.
    The names (keys) are the following:
        'PVs',
        'Tracks', 'MuonTracks',
        'NeutralProtos', 'ChargedProtos',
        'CaloElectrons', 'CaloPhotons', 'CaloMergedPi0s', 'CaloSplitPhotons',
        'MuonPIDs', 'RichPIDs'.
    """
    muonPIDs = reco_unpacker(LocationsPackedReco.PackedMuonPIDs.name,
                             MuonPIDUnpacker, "UnpackMuonPIDs")
    richPIDs = reco_unpacker(
        LocationsPackedReco.PackedRichPIDs.name,
        RichPIDUnpacker,
        "UnpackRichPIDs",
        OutputLevel=ERROR)
    # The OutputLevel above suppresses the following useless warnings (plus more?)
    # WARNING DataPacking::Unpack<LHCb::RichPIDPacker>:: Incorrect data version 0 for packing version > 3. Correcting data to version 2.

    # Ordered so that dependents are unpacked first
    d = collections.OrderedDict([
        ("PVs",
         reco_unpacker(LocationsPackedReco.PackedPVs.name, UnpackRecVertex,
                       "UnpackRecVertices")),
        ("CaloElectrons",
         reco_unpacker(LocationsPackedReco.PackedCaloElectrons.name,
                       UnpackCaloHypo, "UnpackCaloElectrons")),
        ("CaloPhotons",
         reco_unpacker(LocationsPackedReco.PackedCaloPhotons.name,
                       UnpackCaloHypo, "UnpackCaloPhotons")),
        ("CaloMergedPi0s",
         reco_unpacker(LocationsPackedReco.PackedCaloMergedPi0s.name,
                       UnpackCaloHypo, "UnpackCaloMergedPi0s")),
        ("CaloSplitPhotons",
         reco_unpacker(LocationsPackedReco.PackedCaloSplitPhotons.name,
                       UnpackCaloHypo, "UnpackCaloSplitPhotons")),
        ("MuonPIDs", muonPIDs),
        ("RichPIDs", richPIDs),
        ("Tracks",
         reco_unpacker(LocationsPackedReco.PackedTracks.name, UnpackTrack,
                       "UnpackBestTracks")),
        ("MuonTracks",
         reco_unpacker(LocationsPackedReco.PackedMuonTracks.name, UnpackTrack,
                       "UnpackMuonTracks")),
        ("NeutralProtos",
         reco_unpacker(LocationsPackedReco.PackedNeutralProtos.name,
                       UnpackProtoParticle, "UnpackNeutralProtos")),
        ("ChargedProtos",
         reco_unpacker(
             LocationsPackedReco.PackedChargedProtos.name,
             UnpackProtoParticle,
             "UnpackChargedProtos",
             AddInfo=[
                 ChargedProtoParticleAddRichInfo(
                     InputRichPIDLocation=richPIDs.OutputName),
                 ChargedProtoParticleAddMuonInfo(
                     InputMuonPIDLocation=muonPIDs.OutputName),
                 ChargedProtoParticleAddCombineDLLs()
             ])),
    ])

    # Make sure we have consistent names, and that we're unpacking everything
    # we load from the file
    assert set(["Packed" + k for k in d.keys()]) - set(
        enums_as_dict(LocationsPackedReco).keys()) == set()

    return d


def mc_unpackers():
    """
    Return a {object name: `PyConf.Algorithm` instance}  `OrderedDict`
    effectively mapping unpacked reconstruction object names to their respective
    unpacked data.
    The names (keys) are the following:
        'MCRichDigitSummaries',
        'MCParticles', 'MCVertices',
        'MCVPHits', 'MCUTHits', 'MCFTHits','MCRichHits',
        'MCEcalHits', 'MCHcalHits', 'MCMuonHits'.
    """
    # Ordered so that dependents are unpacked first
    mc_vertices = mc_unpacker(LocationsPackedSim.PackedMCVertices.name,
                              UnpackMCVertex, "UnpackMCVertices")
    # Make sure that MC particles and MC vertices are unpacked together,
    # see https://gitlab.cern.ch/lhcb/LHCb/issues/57 for details.
    mc_particles = mc_unpacker(
        LocationsPackedSim.PackedMCParticles.name,
        UnpackMCParticle,
        "UnpackMCParticles",
        ExtraInputs=[mc_vertices])

    mc_vp_hits = mc_unpacker(LocationsPackedSim.PackedMCVPHits.name,
                             UnpackMCVPHit, "UnpackMCVPHits")
    mc_ut_hits = mc_unpacker(LocationsPackedSim.PackedMCUTHits.name,
                             UnpackMCUTHit, "UnpackMCUTHits")
    mc_ft_hits = mc_unpacker(LocationsPackedSim.PackedMCFTHits.name,
                             UnpackMCFTHit, "UnpackMCFTHits")
    mc_rich_hits = mc_unpacker(LocationsPackedSim.PackedMCRichHits.name,
                               UnpackMCRichHit, "UnpackMCRichHits")
    mc_ecal_hits = mc_unpacker(LocationsPackedSim.PackedMCEcalHits.name,
                               UnpackMCEcalHit, "UnpackMCEcalHits")
    mc_hcal_hits = mc_unpacker(LocationsPackedSim.PackedMCHcalHits.name,
                               UnpackMCHcalHit, "UnpackMCHcalHits")
    mc_muon_hits = mc_unpacker(LocationsPackedSim.PackedMCMuonHits.name,
                               UnpackMCMuonHit, "UnpackMCMuonHits")

    mc_rich_digit_sums = mc_unpacker(
        LocationsPackedSim.PackedMCRichDigitSummaries.name, RichSumUnPack,
        "RichSumUnPack")

    d = collections.OrderedDict([
        ("MCRichDigitSummaries", mc_rich_digit_sums),
        ("MCParticles", mc_particles),
        ("MCVertices", mc_vertices),
        ("MCVPHits", mc_vp_hits),
        ("MCUTHits", mc_ut_hits),
        ("MCFTHits", mc_ft_hits),
        ("MCRichHits", mc_rich_hits),
        ("MCEcalHits", mc_ecal_hits),
        ("MCHcalHits", mc_hcal_hits),
        ("MCMuonHits", mc_muon_hits),
    ])

    # Make sure we have consistent names, and that we're unpacking everything
    # we load from the file
    assert set(["Packed" + k for k in d.keys()]) - set(
        enums_as_dict(LocationsPackedSim).keys()) == set()

    return d


def reco_unpacker(key, configurable, name, **kwargs):
    """
    Return a reco. unpacker (`PyConf.Algorithm` instance) that reads from file
    at `LocationsPackedReco[key]` and unpacks to the
    forced output location `LocationsUnpackedReco[key]`.
    """
    alg = Algorithm(
        configurable,
        name=name,
        InputName=make_data_with_FetchDataFromFile(
            LocationsPackedReco[key].value),
        outputs={
            "OutputName": force_location(LocationsUnpackedReco[key].value)
        },
        **kwargs)
    return alg


def mc_unpacker(key, configurable, name, **kwargs):
    """
    Return a sim. unpacker (`PyConf.Algorithm` instance) that reads from file
    at `LocationsPackedSim[key]` and unpacks to the
    forced output location `LocationsUnpackedSim[key]`.
    """
    alg = Algorithm(
        configurable,
        name=name,
        InputName=make_data_with_FetchDataFromFile(
            LocationsPackedSim[key].value),
        outputs={
            "OutputName": force_location(LocationsUnpackedSim[key].value)
        },
        **kwargs)
    return alg


def make_mc_track_info():
    """
    Return the MCTrackInfo data under `locations.LocationMCTrackInfo`
    via `Gaudi::Hive::FetchDataFromFile`.
    """
    return make_data_with_FetchDataFromFile(LocationMCTrackInfo)


def boole_links_digits_mcparticles():
    """
    Return a {TES_path: make_data_with_FetchDataFromFile(TES_path)} dict
    of locations (`locations.LocationsBooleMCParticleLinkers`) for MC linker tables
    (to `MCParticles`) created by Boole.
    """
    return {
        loc.name: make_data_with_FetchDataFromFile(loc.value)
        for loc in LocationsBooleMCParticleLinkers
    }


def boole_links_digits_mchits():
    """
    Return a {TES_path: make_data_with_FetchDataFromFile(TES_path)} dict
    of locations (`locations.LocationsBooleMCHitsLinkers`) for MC linker tables
    (to `MCHits`) created by Boole.

    These locations are only propagated and persisted out of Boole
    for eXtendend DIGI and DST types.
    """
    return {
        loc.name: make_data_with_FetchDataFromFile(loc.value)
        for loc in LocationsBooleMCHitsLinkers
    }


def brunel_links():
    """
    Return a {TES_path: make_data_with_FetchDataFromFile(TES_path)} dict
    of locations (`locations.LocationsBrunelMCLinkers`) for MC linker tables
    created by Brunel.
    """
    return {
        loc.name: make_data_with_FetchDataFromFile(loc.value)
        for loc in LocationsBrunelMCLinkers
    }
