###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Definitions of:

- `Particle` and `ProtoParticle` filters.
- Track selectors.
- Default cuts a la runs 1&2 common particles.
"""
from __future__ import absolute_import, division, print_function

from PyConf.tonic import configurable
from PyConf.Tools import LoKi__Hybrid__ProtoParticleFilter as ProtoParticleFilter
from PyConf.Tools import LoKi__Hybrid__TrackSelector as TrackSelector

from .hacks import patched_hybrid_tool

#########################
# Helpers to combine cuts
#########################


def require_all(*cuts):
    """
    Return a cut string requiring all (string) arguments.

    Example:

        >>> require_all('PT > {pt_min}', 'DLLK < {dllk_max}')
        '(PT > {pt_min}) & (DLLK < {dllk_max})'
    """
    return " & ".join(["({})".format(c) for c in cuts])


def require_any(*cuts):
    """
    Return a cut string requiring at least one of the (string) arguments passes.

    Example:

        >>> require_any('M < 8*GeV', 'PT > 3*GeV')
        '(M < 8*GeV) | (PT > 3*GeV)'
    """
    return " | ".join(["({})".format(c) for c in cuts])


#######################
# Protoparticle filters
#######################


@configurable
def all_protoparticle_filter(Code="PP_ALL", **kwargs):
    """
    Get a `LoKi__Hybrid__ProtoParticleFilter` instance
    that by default selects all protoparticles.

    Args:
        Code (str): The "Code" argument to pass to the filter tool.
                    Default = "PP_ALL".
        kwargs: Keyword arguments accepted by `LoKi__Hybrid__Tool`.

    Returns:
        `LoKi__Hybrid__ProtoParticleFilter` instance wrapped as a `PyConf.components.Tool`.
    """
    return ProtoParticleFilter(
        Code=Code, Factory=patched_hybrid_tool("PPFactory"), **kwargs)


#################
# Track selectors
#################


@configurable
def get_all_track_selector(Code="TrALL", **kwargs):
    """
    Get a `LoKi__Hybrid__TrackSelector` instance
    that by default selects all tracks.

    Args:
        Code (str): The "Code" argument to pass to the tool.
                    Default = "TrALL".
        kwargs: Keyword arguments accepted by `LoKi__Hybrid__TrackSelector`.

    Returns:
        `LoKi__Hybrid__TrackSelector` instance wrapped as a `PyConf.components.Tool`.
    """
    return TrackSelector(Code=Code, **kwargs)


@configurable
def get_long_track_selector(Code='TrALL', **kwargs):
    """
    Get a `LoKi__Hybrid__TrackSelector` instance
    that by default selects all long tracks.

    Args:
        Code (str): The "Code" argument to pass to the tool.
                    Default = "TrALL & TrLONG".
        kwargs: Keyword arguments accepted by `LoKi__Hybrid__TrackSelector`.

    Returns:
        `LoKi__Hybrid__TrackSelector` instance wrapped as a `PyConf.components.Tool`.
    """
    return TrackSelector(Code=require_all("TrLONG", Code), **kwargs)


@configurable
def get_down_track_selector(Code='TrALL', **kwargs):
    """
    Get a `LoKi__Hybrid__TrackSelector` instance
    that by default selects all downstream tracks.

    Args:
        Code (str): The "Code" argument to pass to the tool.
                    Default = "TrALL & TrDOWNSTREAM".
        kwargs: Keyword arguments accepted by `LoKi__Hybrid__TrackSelector`.

    Returns:
        `LoKi__Hybrid__TrackSelector` instance wrapped as a `PyConf.components.Tool`.
    """
    return TrackSelector(Code=require_all("TrDOWNSTREAM", Code), **kwargs)


@configurable
def get_upstream_track_selector(Code='TrALL', **kwargs):
    """
    Get a `LoKi__Hybrid__TrackSelector` instance
    that by default selects all upstream tracks.

    Args:
        Code (str): The "Code" argument to pass to the tool.
                    Default = "TrALL & TrUPSTREAM".
        kwargs: Keyword arguments accepted by `LoKi__Hybrid__TrackSelector`.

    Returns:
        `LoKi__Hybrid__TrackSelector` instance wrapped as a `PyConf.components.Tool`.
    """
    return TrackSelector(Code=require_all("TrUPSTREAM", Code), **kwargs)


#################################
# Default track and particle cuts
#################################


def default_track_cuts():
    """
    Return a string with the default track cuts.
    These are set as a take-all since in principle the track cuts are applied in HLT.
    """
    return require_all("TrALL")


def default_particle_cuts():
    """
    Return a string with the default particle standard loose cuts.
    """
    return require_all("PT>250*MeV", "MIPCHI2DV(PRIMARY)>4.")
