###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Definitions of enums specifying the standard locations of
packed and unpacked objects, and various linker tables.

.. note::
    These locations are what has been used in Runs 1 & 2,
    and may need a revision once the Run 3 event model is finalised
    and the definition of what gets persisted gets formalised.
"""
from __future__ import absolute_import, division, print_function

from enum import Enum


class LocationsPackedReco(Enum):
    """
    Locations of packed reconstruction objects, stored under "/Event/pRec".
    """
    PackedPVs = "/Event/pRec/Vertex/Primary"
    PackedCaloElectrons = "/Event/pRec/Calo/Electrons"
    PackedCaloPhotons = "/Event/pRec/Calo/Photons"
    PackedCaloMergedPi0s = "/Event/pRec/Calo/MergedPi0s"
    PackedCaloSplitPhotons = "/Event/pRec/Calo/SplitPhotons"
    PackedMuonPIDs = "/Event/pRec/Muon/MuonPID"
    PackedRichPIDs = "/Event/pRec/Rich/PIDs"
    PackedTracks = "/Event/pRec/Track/Best"
    PackedMuonTracks = "/Event/pRec/Track/Muon"
    PackedNeutralProtos = "/Event/pRec/ProtoP/Neutrals"
    PackedChargedProtos = "/Event/pRec/ProtoP/Charged"


LocationsUnpackedReco = Enum(
    "LocationsUnpackedReco",
    {e.name: e.value.replace("pRec", "Rec")
     for e in LocationsPackedReco})
LocationsUnpackedReco.__doc__ = """
Locations of unpacked reconstruction objects, stored under "/Event/Rec".
"""


class LocationsPackedSim(Enum):
    """
    Locations of packed simulation objects, stored under "/Event/pSim".
    """
    PackedMCParticles = "/Event/pSim/MCParticles"
    PackedMCVertices = "/Event/pSim/MCVertices"
    PackedMCVPHits = "/Event/pSim/VP/Hits"
    PackedMCUTHits = "/Event/pSim/UT/Hits"
    PackedMCFTHits = "/Event/pSim/FT/Hits"
    PackedMCRichHits = "/Event/pSim/Rich/Hits"
    PackedMCEcalHits = "/Event/pSim/Ecal/Hits"
    PackedMCHcalHits = "/Event/pSim/Hcal/Hits"
    PackedMCMuonHits = "/Event/pSim/Muon/Hits"
    PackedMCRichDigitSummaries = "/Event/pSim/Rich/DigitSummaries"


class LocationsUnpackedSim(Enum):
    """
    Locations of unpacked simulation objects, stored under "/Event/MC".
    """
    PackedMCParticles = "/Event/MC/Particles"
    PackedMCVertices = "/Event/MC/Vertices"
    PackedMCVPHits = "/Event/MC/VP/Hits"
    PackedMCUTHits = "/Event/MC/UT/Hits"
    PackedMCFTHits = "/Event/MC/FT/Hits"
    PackedMCRichHits = "/Event/MC/Rich/Hits"
    PackedMCEcalHits = "/Event/MC/Ecal/Hits"
    PackedMCHcalHits = "/Event/MC/Hcal/Hits"
    PackedMCMuonHits = "/Event/MC/Muon/Hits"
    PackedMCRichDigitSummaries = "/Event/MC/Rich/DigitSummaries"


# Location of MCTrackInfo objects
LocationMCTrackInfo = "/Event/MC/TrackInfo"


class LocationsBooleMCParticleLinkers(Enum):
    """
    Locations of MC linker tables to MCParticles created by Boole.
    """
    EcalDigits = "/Event/Link/Raw/Ecal/Digits"
    FTLiteClusters = "/Event/Link/Raw/FT/LiteClusters"
    HcalDigits = "/Event/Link/Raw/Hcal/Digits"
    MuonDigits = "/Event/Link/Raw/Muon/Digits"
    UTClusters = "/Event/Link/Raw/UT/Clusters"
    VPDigits = "/Event/Link/Raw/VP/Digits"


class LocationsBooleMCHitsLinkers(Enum):
    """
    Locations for MC linker tables to MCHits created by Boole.

    These locations are only propagated out of Boole for eXtendend DIGI and DST types.
    """
    FTLiteClusters = "/Event/Link/Raw/FT/LiteClusters2MCHits"
    UTClusters = "/Event/Link/Raw/UT/Clusters2MCHits"
    VPDigits = "/Event/Link/Raw/VP/Digits2MCHits"


class LocationsBrunelMCLinkers(Enum):
    """
    Locations of MC linker tables created by Brunel.
    """
    CaloElectrons = "/Event/Link/Rec/Calo/Electrons"
    CaloMergedPi0s = "/Event/Link/Rec/Calo/MergedPi0s"
    CaloPhotons = "/Event/Link/Rec/Calo/Photons"
    CaloSplitPhotons = "/Event/Link/Rec/Calo/SplitPhotons"
    Tracks = "/Event/Link/Rec/Track/Best"


def enums_as_dict(enums):
    """
    Return a {name: value} dict of all enum members.

    Example:

        >>> class MyEnum(Enum):
        ...     a = 1
        ...     b = 2
        >>> enums_as_dict(MyEnum)
        {'a': 1, 'b': 2}
    """
    return {e.name: e.value for e in enums}
