###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Define value and run-time errors used in DaVinci application.
"""
import click


class DVOptionError(ValueError):
    """
    DaVinci exception class for invalid option configurations.
    """

    def __init__(self, name, value):
        """
        Constructor.
        """
        self.message = f"invalid DaVinci '{name}' option value specified: '{value}'"

    def __str__(self):
        """
        String representation.
        """
        return f"{click.style('DVOptionError', bold=True, fg='red')} {click.style(self.message, fg='red')}"


class DVRuntimeError(RuntimeError):
    """
    DaVinci exception class for inconsistent option configurations.
    """

    def __init__(self, option1, option2, value):
        """
        Constructor.
        """
        self.message = f"Option {option1} and {option2} are inconsistent! \n {value}"

    def __str__(self):
        """
        String representation.
        """
        return f"{click.style('DVRuntimeError', bold=True, fg='red')} {click.style(self.message, fg='red')}"


class DVImportError(ImportError):
    """
    DaVinci exception class for an invalid Python module import.
    """

    def __init__(self, algName, value):
        """
        Constructor.
        """
        self.message = f"Importing user algorithm '{algName}' failed! \n {value}"

    def __str__(self):
        """
        String representation.
        """
        return f"{click.style('DVImportError', bold=True, fg='red')} {click.style(self.message, fg='red')}"


class DVKeyError(KeyError):
    """
    DaVinci exception class for accessing to an invalid key option.
    """

    def __init__(self, optName):
        """
        Constructor.
        """
        self.message = f"There is no option named '{optName}' in DaVinci!"

    def __str__(self):
        """
        String representation.
        """
        return f"{click.style('DVKeyError', bold=True, fg='red')} {click.style(self.message, fg='red')}"


def option_checker(name, value):
    """
    Check the correctness of a DaVinci configurable option.
    """
    from DaVinci.options_default import __optsDict__ as dv_default

    allowedValues = get_allowed_option_values()
    if name in dv_default.keys() and name not in allowedValues.keys():
        log_click(
            "INFO",
            f"The option {name} has no restrinctions on the allowed values.")
    else:
        try:
            if value not in allowedValues[name]:
                print_allowed_option_values(allowedValues, name)
                raise DVOptionError(name, value)
        except KeyError:
            log_click("ERROR",
                      f"You are trying to check the unknown option {name}!")
            print_allowed_option_values(allowedValues)
            raise DVKeyError(name) from None


def get_allowed_option_values():
    allowedValues = {
        "data_type": ["Upgrade"],
        "detectors": [
            'VP', 'UT', 'FT', 'Rich1Pmt', 'Rich2Pmt', 'Ecal', 'Hcal', 'Muon',
            'Magnet', 'Tr'
        ],
        "input_type": [
            "MDF", "DST", "DIGI", "RDST", "MDST", "SDST", "XDST", "LDST",
            "ROOT", "XGEN"
        ]
    }
    return allowedValues


def print_allowed_option_values(allowedValues, name=None):
    """
    Print the allowed values for the DaVinci option configurations.
    """
    if name is None:
        log_click("", "Known job option configurations and allowed values:",
                  "green")
        for name, values in allowedValues.items():
            log_click("", f"{name} \t : {values}")
    else:
        log_click("", f"Allowed values for DaVinci option {name}:", "green")
        log_click("", f"{allowedValues[name]}")


def log_click(mode, message, color_text="white"):
    import click
    config_colors = {
        "INFO": "blue",
        "WARNING": "yellow",
        "ERROR": "red",
    }

    color_mode = config_colors[mode] if mode else "white"
    click.echo(
        f"{click.style(mode, bold=True, fg=color_mode)} {click.style(message, fg=color_text)}"
    )
