##############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Configurables to make the various reconstruction objects from the
packed data on file.

.. note::
    1) What is defined here relies on data paths used in Runs 1 & 2,
    and may need a revision once the Run 3 event model is finalised
    and the definition of what gets persisted gets formalised.
    2) Code very heavily relies on its Moore equivalent. Thank you, RTA team.
"""
from GaudiConf import PersistRecoConf
from PyConf import configurable
from PyConf.Algorithms import (
    LHCb__Converters__RecVertices__LHCbRecVerticesToVectorV2RecVertex,
    LHCb__Converters__Track__v2__fromV1TrackV2Track)
from .data_from_file import reco_unpackers


def upfront_reconstruction_from_file():  # renamed from upfront_reconstruction
    """Return a list DataHandles that define the upfront reconstruction output.

    This differs from `reconstruction` as it should not be used as inputs to
    other algorithms, but only to define the control flow, i.e. the return
    value of this function should be ran before all HLT2 lines.

    """
    return list(reco_unpackers().values())


def reconstruction_from_file():  # renamed from reconstruction
    """Return a {name: DataHandle} dict that define the reconstruction output."""
    return {k: v.OutputName for k, v in reco_unpackers().items()}


def make_charged_protoparticles():
    return reconstruction()['ChargedProtos']


def make_neutral_protoparticles():
    return reconstruction()['NeutralProtos']


def make_pvs():
    return reconstruction()['PVs']


def make_tracks():
    return reconstruction()['Tracks']


def make_pvs_using(tracks, pvs):
    tracks_v2 = LHCb__Converters__Track__v2__fromV1TrackV2Track(
        InputTracksName=tracks).OutputTracksName
    return LHCb__Converters__RecVertices__LHCbRecVerticesToVectorV2RecVertex(
        InputVertices=pvs,
        InputTracks=tracks_v2,
    ).OutputVertices


def make_pvs_with(persist_conf, tracks='Tracks', pvs='PVs'):
    return make_pvs_using(persist_conf.unpackers_by_key()[tracks],
                          persist_conf.unpackers_by_key()[pvs])


def make_pvs_for(process="Spruce", data_type="Upgrade"):
    if process == 'Spruce':
        stream_pvs = 'Spruce/HLT2'
    else:
        stream_pvs = 'HLT2'
    conf = PersistRecoConf.PersistRecoPacking(
        stream='/Event/' + stream_pvs, data_type=data_type)
    return make_pvs_with(conf)


@configurable
def reconstruction(from_file=True):
    """Return reconstruction objects.

    Note:
        It is advised to use this function if more than one object is needed,
    rather than the accessors below as it makes the configuration slower.
    """
    # removed reco since it will not be done in DV
    reco = reconstruction_from_file()
    upfront_reconstruction = upfront_reconstruction_from_file()

    charged_protos = reco["ChargedProtos"]
    neutral_protos = reco["NeutralProtos"]
    best_tracks = reco["Tracks"]
    pvs = reco["PVs"]
    electrons = reco["CaloElectrons"]
    photons = reco["CaloPhotons"]
    mergedPi0s = reco["CaloMergedPi0s"]
    splitPhotons = reco["CaloSplitPhotons"]
    muon_pids = reco["MuonPIDs"]
    rich_pids = reco["RichPIDs"]

    return {
        "ChargedProtos": charged_protos,
        "NeutralProtos": neutral_protos,
        "Tracks": best_tracks,
        "PVs": pvs,
        "UpfrontReconstruction": upfront_reconstruction,
        "CaloElectrons": electrons,
        "CaloPhotons": photons,
        "CaloMergedPi0s": mergedPi0s,
        "CaloSplitPhotons": splitPhotons,
        "MuonPIDs": muon_pids,
        "RichPIDs": rich_pids,
    }
