###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
This module contains some functions useful while running the davinci script.
"""


def dump_call(testfiledb_file, testfiledb_key, joboptfile, ctx_args):
    """
    Print out all the davinci options used to run the job by means of the 'davinci' script.

    Args:
       - testfiledb_file: TestFileDB-like file with job input information (1st davinci argument, 1st slot).
       - testfiledb_key: relevant key to be looked for in the TestFileDB-like file (1st davinci argument, 2nd slot).
       - joboptfile: file containing the job option information (2nd davinci argument).
       - ctx_args: array with the remnant options passed to davinci by command line.
    """
    import click

    indentStr = "# "
    click.echo("%sDaVinci: running using the following arguments:" % indentStr)
    click.echo(
        "%s         - TestFileDB file: %s" % (indentStr, testfiledb_file))
    click.echo(
        "%s         - TestFileDB key:  %s" % (indentStr, testfiledb_key))
    click.echo("%s         - joboptfile:  %s" % (indentStr, joboptfile))

    if len(ctx_args) > 1:
        for i in range(0, len(ctx_args), 2):
            click.echo("%s         - %s: %s" % (indentStr, ctx_args[i][2:],
                                                ctx_args[i + 1]))


def get_configurable_opts(configurables, with_defaults):
    """
    Loop over the list of configurables to be passed to Gaudi and convert them
    in a dictionary.

    Args:
       - configurables: list of configurables to be passed to Gaudi.
       - with_defaults: flag to set the default values.

    Returns:
       - opts: dictionary containing all the configurables.
    """
    import sys
    from itertools import chain
    from GaudiKernel.Proxy.Configurable import Configurable

    opts = {}
    for c in configurables:
        if hasattr(c, "__opt_properties__"):
            opts.update(c.__opt_properties__(with_defaults))
        else:
            items = (chain(c.getDefaultProperties().items(),
                           c.getValuedProperties().items())
                     if with_defaults else c.getValuedProperties().items())

            for p, v in items:
                if hasattr(Configurable, "PropertyReference") and isinstance(
                        v, Configurable.PropertyReference):
                    v = v.__resolve__()
                if isinstance(v, str):
                    v = '"%s"' % v.replace('"', '\\"')
                elif sys.version_info < (
                        3,
                ):  # TODO: remove as soon as Python 2 gets unsupported
                    import types
                    if type(v) is types.LongType:
                        v = '%d' % v
                elif hasattr(v, '__opt_value__'):
                    v = v.__opt_value__()

                opts['.'.join((c.name(), p))] = str(v)
    return opts


def set_testfiledb(inputfiledb):
    """
    Set the TestFileDB file and the related key to be used in the job.

    Args:
        inputfiledb (str): pair of values containing information on the TestFileDB and the key chosen by the user.

    Returns:
        key for the TestFileDB.
        file containing the TestFileDB.
    """
    assert len(
        inputfiledb
    ) in (1, 2), "--inputfiledb takes two arguments: filedb key and the related filename."\
        "If the filename is not specified the TestFileDB will be used as default. %d" %len(inputfiledb)

    inputfiledb_key = inputfiledb[0]
    inputfiledb_file = inputfiledb[
        1] if not inputfiledb[1] == "-" else "TestFileDB"

    return inputfiledb_key, inputfiledb_file


def get_dummy_config():
    """
    Simple function returning a dummy instance of the configurables to be passed to DaVinci.
    """
    from PyConf.application import ComponentConfig
    from Configurables import ApplicationMgr

    config = ComponentConfig()
    config.add(ApplicationMgr(AppName="DaVinci"))

    return config


def create_jobopt_template(file_name="jobopt_template.yaml",
                           interactive=False):
    """
    Function to create a simple template file with job options
    as specified by the --joboptfile davinci script option.

    Args:
        file_name (str): name for the output template file. Default = 'jobopt_template.yaml'.
        interactive (bool): flag to create the template interactively.

    Output:
        the template file with job options and the name defined by the user.
    """
    if not interactive:
        template_str = "# Template job option YAML file.\n"\
                       "# Best guesses are provided below for various options. Please adapt as per your needs.\n\n"\
                       "annsvc_config: '<path_to_annsvc_json>'\n"\
                       "evt_max: -1\n"\
                       "ntuple_file: 'ntuple_file.root'\n"\
                       "histo_file: 'histos_file.root'\n"\
                       "enable_unpack: True\n"\
                       "process: 'Spruce'\n"

        with open(file_name, 'w') as jobopt_file_file:
            jobopt_file_file.write(template_str)
    else:
        import yaml, click
        # Get the DaVinci dict containing the default values.
        from DaVinci.options_default import __optsDict__ as opts_dict

        dict_template = {}
        # These are the data options that are set via the inputfiledb, so we remove them from the joboptfile.
        list_data_qualifiers = [
            "data_type", "input_type", "simulation", "conddb_tag", "dddb_tag",
            "input_files"
        ]

        for name, config in opts_dict.items():
            if name not in list_data_qualifiers:
                value = click.prompt(
                    f"Set the value for the option {name}:\n[{config['text']}]\n",
                    default=config["value"],
                    type=type(config["value"]))
                # Adding to the dict only if the value is not the default one.
                if not value == config["value"]:
                    dict_template.update({name: value})

        if len(dict_template):
            with open(file_name, 'w') as jobopt_file:
                yaml.safe_dump(
                    dict_template, jobopt_file, default_flow_style=False)
        else:
            click.echo(
                "All option values match the default ones, no job option file is needed!"
            )


def create_inputdb_template(file_name="inputdb_template.yaml"):
    """
    Function to create a simple template file for the TestFileDB-like inputdb file.

    Args:
        file_name (str): name for the output template file. Default = 'inputdb_template.yaml'.

    Output:
        the template TestFileDB-like file with input data options and the name defined by the user.
    """

    template_str = "# Template input data YAML file.\n"\
                   "# Best guesses are provided below for various qualifiers. Please adapt as per your needs.\n\n"\
                   "MyInputData:\n"\
                   "   filenames:\n"\
                   "      - '<a_file_path>'\n"\
                   "      - '<another_file_path>'\n"\
                   "   qualifiers:\n"\
                   "      data_type: Upgrade\n"\
                   "      input_type: DST\n"\
                   "      simulation: true\n"\
                   "      conddb_tag: <e.g. sim-20171127-vc-md100>\n"\
                   "      dddb_tag: <e.g. dddb-20171126>\n"\
                   "   metadata:\n"\
                   "      Author: '<me :-)>'\n"\
                   "      Date: '<date_of_creation>'\n"\
                   "      Comment: '<E.g. how the files were created or for what purpose>'"

    with open(file_name, 'w') as inputdb_file:
        inputdb_file.write(template_str)
