###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest, os

from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer

from DaVinci import options
from DaVinci.optionChecker import DVImportError
from DaVinci.algorithms import (setup_algorithms, define_fsr_writer,
                                add_filter, apply_filters_and_unpacking,
                                setup_user_algorithms, unpack_locations,
                                configured_FunTuple)


def test_import_main_options():
    """
    Check if a configurable .opts file can be imported correctly in DaVinci.
    """
    filename = "test_options.opts"
    with open(filename, "w") as f:
        f.write("DVAppOptions.enable_unpack = True;\n")
        f.write("DVAppOptions.evt_max = 5;\n")
        f.close()

    options.main_options = filename
    options.user_algorithms = ""
    setup_algorithms(options)
    os.remove(filename)

    # Restore default value avoid failure in the next tests
    options.main_options = ""


def test_set_user_algorithms():
    """
    Check if DaVinci can import correctly an external user algorithm implemented in a given Python module.
    """
    filename = "test_useralgs"
    with open(f"{filename}.py", "w") as f:
        f.write("""
from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
from DaVinci import options

options.data_type = 'Upgrade'
options.enable_unpack = False
options.evt_max = 10

def main():
    algs = {'testAlg': [VoidConsumer()]}
    return algs, []
""")

    options.user_algorithms = f"{filename}:main"
    test_algs, _ = setup_algorithms(options)
    assert "testAlg" in test_algs

    os.remove(f"{filename}.py")

    # Restore default value avoid failure in the next tests
    options.user_algorithms = ""


def test_no_user_algorithms():
    """
    Check that no algorithms are imported if user doens't provide a python module.
    """
    test_algs, _ = setup_user_algorithms("")
    assert len(test_algs) == 0


def test_user_algorithms_with_ext():
    """
    Check that a Python module can be imported correctly
    even if the '.py' extension is passed by mistake by the user.
    """
    filename = "./test_useralg.py"
    with open(filename, "w") as f:
        f.write("""
from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
from DaVinci import options

options.data_type = 'Upgrade'
options.enable_unpack = False
options.evt_max = 10

def main():
    algs = [VoidConsumer()]
    return algs, []
""")

    test_algs, _ = setup_user_algorithms(f"{filename}:main")
    assert "UserAlgorithms" in test_algs

    os.remove(filename)


def test_user_algorithms_no_module():
    """
    Check DaVinci returns the correct error if the Python module defined by the user doesn't exist.
    """
    filename = "non_existent_test_useralg"
    with pytest.raises(DVImportError) as excinfo:
        test_algs, _ = setup_user_algorithms(f"{filename}.py")

    assert "Importing user algorithm '' failed!" in str(excinfo.value)
    assert f"Check why the user Python module '{filename}' defined in './' can't be imported!" in str(
        excinfo.value)


def test_user_algorithms_no_function():
    """
    Check DaVinci returns the correct error if the user provides only the module
    and not the name of the function to be run.
    """
    filename = "test_useralg"
    with open(f"{filename}.py", "w") as f:
        f.write("""
rom PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
from DaVinci import options

options.data_type = 'Upgrade'
options.enable_unpack = False
options.evt_max = 10
""")

    with pytest.raises(TypeError):
        test_algs, _ = setup_user_algorithms(filename)

    os.remove(f"{filename}.py")


def test_user_algorithms_wrong_function():
    """
    Check DaVinci returns the correct error if the function specified by the user doesn't exist.
    """
    filename = "test_useralg"
    with open(f"{filename}.py", "w") as f:
        f.write("""
from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
from DaVinci import options

options.data_type = 'Upgrade'
options.enable_unpack = False
options.evt_max = 10

def main():
    algs = {'testAlg': [VoidConsumer()]}
    return algs, []
""")

    with pytest.raises(AttributeError):
        test_algs, _ = setup_user_algorithms(f"{filename}:main2")

    os.remove(f"{filename}.py")


def test_define_write_fsr():
    """
    Check if DaVinci imports correctly the algorithm to merge and write FSRs.
    """
    options.output_level = 3
    options.merge_genfsr = True
    options.simulation = True
    test_algs = define_fsr_writer(options)
    assert any("GenFSRMerge" == x.name for x in test_algs)


def test_add_hlt2_filter():
    """
    Check if DaVinci is able to implement correctly a filter on an HLT2 line.
    """
    options.process = "Hlt2"
    options.stream = "default"
    test_filter = add_filter("test_filter", "HLT_PASS('Hlt2TESTLineDecision')")
    assert "HDRFilter" in test_filter.fullname


def test_add_spruce_filter():
    """
    Check if DaVinci is able to implement correctly a filter on a Sprucing line.
    """
    options.process = "Spruce"
    options.stream = "default"
    test_filter = add_filter("test_filter",
                             "HLT_PASS('SpruceTESTLineDecision')")
    assert "HDRFilter" in test_filter.fullname


def test_add_void_filter():
    """
    Check if DaVinci is able to implement correcty a Void filter
    if 'HLT_PASS' string is not found in the filter code."
    """
    test_filter = add_filter("test_filter", "VOIDTEST_Filter")
    assert "VoidFilter" in test_filter.fullname


def test_unpack_locations():
    """
    Check if the unpacking algorithms are retrieved correctly from GaudiConf.
    TO BE REMOVED WHEN THE UNPACKING WILL BECOME FULLY FUNCTIONAL.
    """
    options.simulation = True
    test_algs = unpack_locations(options, False)

    assert isinstance(test_algs, list)
    assert any("Unpack" in alg.fullname for alg in test_algs)


def test_unpack_locations_xgen():
    """
    Check if DaVinci unpacks only the MC locations if run on an xgen file.
    (i.e. 'unpack_only_mc' is set to True.)
    TO BE REMOVED WHEN THE UNPACKING WILL BECOME FULLY FUNCTIONAL.
    """
    options.simulation = True
    unpack_only_mc = True
    test_algs = unpack_locations(options, unpack_only_mc)

    assert isinstance(test_algs, list)
    assert any("UnpackMCParticle" in alg.fullname for alg in test_algs)
    assert not any("UnpackTrack" in alg.fullname for alg in test_algs)


def test_apply_filters_and_unpack():
    """
    Check if DaVinci applies correctly a filter in front of a given algorithm
    and instantiates correctly the main unpacking algorithms.
    TO BE UPDATED WHEN THE UNPACKING WILL BECOME FULLY FUNCTIONAL.
    """
    options.evt_pre_filters = {"test_filter": "EVT_PREFILTER"}
    options.enable_unpack = True
    alg_dict = {"test_alg": [VoidConsumer()]}
    test_alg_dict = apply_filters_and_unpacking(options, alg_dict, False)
    list_of_main_expected_algs = [
        "LoKi__VoidFilter", "LHCb__UnpackRawEvent", "HltPackedDataDecoder",
        "UnpackMCParticle", "UnpackTrack"
    ]

    for exp_alg in list_of_main_expected_algs:
        assert any(
            exp_alg in alg.fullname for alg in test_alg_dict["test_alg"])


def test_configured_funtuple():
    """
    Check if the configured_FunTuple provides a correct instance of FunTuple.
    """
    from FunTuple.functorcollections import Kinematics
    fields = {'B0': "[B0 -> D_s- pi+]CC"}
    variables = {'B0': Kinematics()}
    config = {
        "TestTuple": {
            "location": "/Event/Spruce/SpruceTestLine/Particles",
            "filters":
            ["HLT_PASS('SpruceTestLine1Decision')", "VoidTest_Filter"],
            "preamble": ['TRACK_MAX_PT = MAXTREE(ISBASIC & HASTRACK, PT, -1)'],
            "tuple": "DecayTree",
            "fields": fields,
            "variables": variables,
        }
    }

    test_dict = configured_FunTuple(config)
    assert any("FunTupleBase_Particles/Tuple_TestTuple" in alg.fullname
               for alg in test_dict["TestTuple"])
