###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest

from DaVinci.optionChecker import DVOptionError, DVRuntimeError
from DaVinci.configOptions import (list_data_options, set_option_value,
                                   set_input_file_options, set_job_options,
                                   check_options, is_option_settable)
from DaVinci import options
from pathlib import Path
DIR = Path(__file__).parent.resolve()

fileDB_file = DIR / "../../options/DaVinciDB-Example.yaml"
fileDB_key = "Upgrade_Bd2KstarMuMu_ldst"
fileDB = "TestFileDB"
jobOpt_file = DIR / "../../options/jobOptions-Example"


def test_list_data_options_key_in_TestFileDB():
    """
    Check if the list of data options in TestFileDB matches the one expected by DaVinci.
    """
    key = "Upgrade_Bd2KstarMuMu"
    assert list_data_options(key, fileDB) == [
        'Author', 'Format', 'DataType', 'Date', 'Simulation', 'CondDB', 'DDDB'
    ]


def test_list_data_options_key_in_userDB():
    """
    Check if the list of data options in a TestFileDB-like file prepared as example is updated.
    """
    assert list_data_options(fileDB_key, fileDB_file) == [
        'data_type', 'input_type', 'simulation', 'conddb_tag', 'dddb_tag'
    ]


def test_list_data_options_unknown_key_in_TestFileDB():
    """
    Check if DaVinci returns an error if the key used for the TestFileDB is unknown.
    """
    with pytest.raises(KeyError) as excinfo:
        key = "UnknownKey"
        _ = list_data_options(key, fileDB)

    assert "UnknownKey" in str(excinfo.value)


def test_handle_options():
    """
    Check if DaVinci is able to set correctly the value of a given option.
    """
    opt_name = "ntuple_file"
    opt_value = "ntupleDV_test.root"
    set_option_value(options, opt_name, opt_value)
    assert options.getProp(opt_name) == opt_value


def test_option_correct_str2bool_conversion():
    """
    Check if DaVinci converts correctly the value of a boolean option even if the value is passed as string.
    """
    opt_name = "enable_unpack"
    opt_value = "True"
    set_option_value(options, opt_name, opt_value)
    assert options.getProp(opt_name)


def test_option_wrong_str2bool_conversion():
    """
    Check if DaVinci raises an error if the string-to-bool conversion is applied on a generic string content.
    """
    opt_name = "enable_unpack"
    opt_value = "unknown_value"

    with pytest.raises(DVOptionError) as excinfo:
        set_option_value(options, opt_name, opt_value)

    assert f"invalid DaVinci '{opt_name}' option value specified:" in str(
        excinfo.value)


def test_set_input_file_from_TestFileDB():
    """
    Check if DaVinci is able to set correctly input data options given a key and the TestFileDB.
    """
    key = "Upgrade_Bd2KstarMuMu"
    set_input_file_options(options, key, fileDB)
    assert not options.input_files == []


def test_set_input_file_from_userDB():
    """
    Check if DaVinci is able to set correctly input data options 
    given a key and TestFileDB-like file database defined by the user.
    """
    set_input_file_options(options, fileDB_key, fileDB_file)
    assert not options.input_files == []


def test_select_input_file_by_index():
    """
    Check if DaVinci selects only a specific file from the ones listed in the database
    if the file index is provided after the database key.
    """
    key = f"{fileDB_key}:3"
    set_input_file_options(options, key, fileDB_file)
    assert len(options.input_files) == 1


def test_wrong_input_file_index():
    """
    Check that DaVinci raises an error if the file index is higher than the number of 
    input files available in the TestFileDB-like database.
    """
    key = f"{fileDB_key}:100"
    with pytest.raises(ValueError):
        set_input_file_options(options, key, fileDB_file)


def test_job_options_from_py_file():
    """
    Check if DaVinci sets correctly an option value when reading a .py job option file.
    Check also the .py example is well formatted and readable by DaVinci.
    """
    set_job_options(options, f"{jobOpt_file}.py", "", "", False)
    assert options.evt_max == 100


def test_job_options_from_yaml_file():
    """
    Check if DaVinci sets correctly an option value when reading a .yaml job option file.
    Check also the .yaml example is well formatted and readable by DaVinci.
    """
    set_job_options(options, f"{jobOpt_file}.yaml", "", "", False)
    assert options.evt_max == 100


def test_job_options_from_unknown_file():
    """
    Check that DaVinci raises an error if the job option file passed by the user doesn't exist.
    """
    import os
    job_opt_name = "test_job_options.txt"
    with open(job_opt_name, "w") as f:
        f.write("evt_max: 100")
        f.close()
    with pytest.raises(ValueError):
        set_job_options(options, job_opt_name, "", "", False)
    os.remove(job_opt_name)


def test_lumi_fails_on_MC():
    """
    Check if enabling lumi counters fails on simulated input files.
    """
    options.simulation = True
    options.lumi = True

    with pytest.raises(DVRuntimeError) as excinfo:
        check_options(options)

    assert "Option lumi and simulation are inconsistent!" in str(excinfo.value)
    assert "lumi not valid for simulation! Set lumi = False." in str(
        excinfo.value)

    # Restore default value avoid failure in the next tests
    options.lumi = False


def test_merge_genfsr_fails_on_data():
    """
    Check that DaVinci raises a runtime error when trying to merge genFSRs on real data.
    """
    options.simulation = False
    options.merge_genfsr = True
    with pytest.raises(DVRuntimeError) as excinfo:
        check_options(options)

    assert "Option merge_genfsr and simulation are inconsistent!" in str(
        excinfo.value)
    assert "GenFSR are not available in real data so merge_genfsr cannot be run! Set merge_genfsr = False." in str(
        excinfo.value)

    # Restore default value avoid failure in the next tests
    options.simulation = True
    options.merge_genfsr = False


def test_mcJob_fails_without_dddb():
    """
    Check that DaVinci job fails on simulated data if no dddb_tag condition is set.
    """
    options.dddb_tag = ""
    with pytest.raises(DVOptionError):
        check_options(options)


def test_mcJob_fails_without_conddb():
    """
    Check that DaVinci job fails on simulated data if no conddb_tag condition is set. 
    """
    options.dddb_tag = "dddb-20171126"
    options.conddb_tag = ""
    with pytest.raises(DVOptionError):
        check_options(options)
    # Restore default value avoid failure in the next tests
    options.dddb_tag = ""


def test_set_data_option_without_override():
    """
    Check that DaVinci raises an error if the user tries to change the value of an input data option
    without explicitely enabling the '--override' flag.
    """
    override_data_options = False
    data_options = list_data_options(fileDB_key, fileDB_file)
    with pytest.raises(DVRuntimeError):
        is_option_settable(options, "data_type", data_options,
                           override_data_options)


def test_set_data_option_with_override():
    """
    Check that DaVinci updates the value of an input data option if the '--override' flag is enabled.
    """
    override_data_options = True
    data_options = list_data_options(fileDB_key, fileDB_file)
    assert is_option_settable(options, "data_type", data_options,
                              override_data_options)
