###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest

from PyConf.Algorithms import Gaudi__Examples__VoidConsumer as VoidConsumer
from PyConf.Algorithms import GaudiHistoAlgorithm

from DaVinci.Configuration import run_davinci_app
from DaVinci.config import DVNode, prepare_davinci_nodes, run_davinci
from DaVinci import options

from pathlib import Path
DIR = Path(__file__).parent.resolve()


def test_run_app_dummy():
    """
    Check if DaVinci returns a ComponentConfig object even when passing only the input data.
    """
    from PyConf.application import ComponentConfig
    configurables = run_davinci_app(
        fileDB_key="Upgrade_Bd2KstarMuMu", fileDB_file="TestFileDB")
    assert isinstance(configurables, ComponentConfig)


def test_run_simple_job():
    """
    Check if DaVinci is able to set correctly the control flow application when running a simple job.
    """
    config = run_davinci_app(
        fileDB_key="Upgrade_Bd2KstarMuMu",
        fileDB_file="TestFileDB",
        jobOpt_file=DIR / "../../options/jobOptions-Example.yaml",
        simple_job=True)
    assert "HLTControlFlowMgr/HLTControlFlowMgr" in config


def test_set_hltAnnSvc():
    """
    Check if DaVinci is able to set correctly the HltAnnSvc given a specific .tck.json file.
    """
    config = run_davinci_app(
        fileDB_key="Upgrade_Bd2KstarMuMu",
        fileDB_file="TestFileDB",
        ctx_args=[
            "--annsvc_config",
            "root://eoslhcb.cern.ch//eos/lhcb/wg/dpa/wp3/tests/spruce_all_lines_realtime.tck.json"
        ])
    assert "HltANNSvc/HltANNSvc" in config


def test_lumi_counters_in_data():
    """
    Check that Lumi Counters are correctly added in the control flow when running on real data.
    """
    config = run_davinci_app(
        fileDB_key="Upgrade_Bd2KstarMuMu",
        fileDB_file="TestFileDB",
        ctx_args=[
            "--simulation", "False", "--write_fsr", "True", "--lumi", "True"
        ],
        override_data_options=True)
    nodes = config["HLTControlFlowMgr/HLTControlFlowMgr"].CompositeCFNodes
    # Node of the HLTControlFlowMgr are tuple containing : (<node_name>, <logic>, list_of_algs, <force_order>)
    # E.g. : ('DaVinci', 'LAZY_AND', ['FileSummaryRecords', 'UserAnalysis'], True),
    assert any("EventAccounting/EventAccount" in algs for algs in nodes[2])


def test_DVNode():
    """
    Basic test of the DVNode class.
    """
    node = DVNode("MyAlgs", [VoidConsumer()])

    assert node.name == "MyAlgs"


def test_prepare_davinci_nodes():
    """
    Preparation of a minimalistic node with a single algorithm (has to be passed as a list).
    """
    list_of_algs = {"alg": [VoidConsumer()]}
    prepare_davinci_nodes(list_of_algs)


def test_prepare_davinci_nodes_TypeError():
    """
    Check that DaVinci raises an error when trying to prepare a node with a set of algorithms different from a list.
    """
    list_of_algs = {"alg": VoidConsumer()}
    with pytest.raises(TypeError):
        prepare_davinci_nodes(list_of_algs)


def test_run_davinci():
    """
    Check that a simple job run similary to run_moore(...) is set-up correctly
    (would be run as gaudirun.py ...).
    """
    # Set the database up
    options.evt_max = 7
    options.set_input_and_conds_from_testfiledb('Upgrade_Bd2KstarMuMu')

    simple_histos = GaudiHistoAlgorithm(name="SimpleHistos", HistoPrint=True)

    # Run davinci
    run_davinci(options, user_algs={"DVselection": [simple_histos]})
